c-----------------------------------------------------------------------
      program surfgen
c-----------------------------------------------------------------------
c
c --- SURFGEN   Version: 7.0.0          Level: 150330              MAIN
c               D. Strimaitis, Exponent Inc.
c
c --- PURPOSE:  Read 1 or more files containing meteorological data
c               for 1 station in each file, extract subset of these
c               data, and create one SURF.DAT file of surface
c               meteorological data and/or one PRECIP.DAT file of
c               precipitation data containing these data for all
c               stations.  Input file types include:
c                  NCDC Integrated Surface Data (ISD) -- hourly
c                  ASOS1 subhourly
c                  Generic CSV format
c
c --- UPDATES:
c               
c  EXTERNAL ROUTINES:   GETCMD, CONTROL, SETCODES, SETUP
c                       COMP, FINISH, OPEN_ERR
c-----------------------------------------------------------------------
c --- Access module data
      use userinf

      implicit none

      integer :: ictrl,iolst,iocode,iosurf,ioprec,iomet,iox
      parameter(ictrl=10, iolst=11, iocode=13, iosurf=14, ioprec=15,
     &          iox=99,iomet=1000)

      integer :: i, ierr

c --- Unit numbers are defined as parameters
c --- IOMET must be greater than all other file unit parameters

c --- File unit numbers for input met files
c ---    NFILES filenames use units starting with iomet+1
c ---    ASOS1 types require 2 files, and the unit for the second
c ---    filename uses iomet*10 as the base
c ---    Example:  input met 4 is type ASOS1, requiring 2 files
c ---              cmet6405 is opened on unit = iomet+4
c ---              cmet6406 is opened on unit = iomet*10+4
c ---    Example:  input met 11 is type ISD, requiring 1 file
c ---              cmetfile is opened on unit = iomet+11

      character(len=256)  :: cfname

c --- Control file name (default or specified on command line)
      cfname='SURFGEN.INP'
      call GETCMD(cfname)

c --- Open and process control file (opens list file also)
      open(ictrl,file=cfname,status='old',iostat=ierr)
      if(ierr.NE.0) call OPEN_ERR(-1,'(main)','Control File',
     &                            cfname,ictrl)
      call CONTROL(ictrl,iolst)
      close(ictrl)

c --- Check for adequate number of file unit numbers for input met
      do i=1,nfiles
         if(ctype(i).EQ.'ASOS1   ') then
            if((iomet+nfiles).GE.(10*iomet+i)) then
c ---          Second ASOS1 file unit number is not unique
               write(iolst,*)'ERROR: Unit number for ASOS1 6406 '//
     &                       'file is not unique ',(10*iomet+i)
               write(iolst,*)'       Unit numbers for ASOS1 6405 '//
     &                       'files extent to ',(iomet+nfiles)
               stop 'HALT in (main) -- see list file'
            endif
c ---       Check is needed only on first ASOS1 file entry
            EXIT
         endif
      enddo

c --- Set up ISD field codes
      open(iocode,file=cisd_code,status='old',iostat=ierr)
      if(ierr.NE.0) call OPEN_ERR(iolst,'(main)','ISD field codes',
     &                            cisd_code,iocode)
      call SETCODES(iocode,iolst)
      close(iocode)

c --- Preparation
      call SETUP(iomet,iolst,iosurf,ioprec)

c --- Process all input files
      call COMP(iomet,iolst,iosurf,ioprec)

c --- Finish
      call FINISH(iolst)

      stop
      end

c --- Utilities
      include 'utils.for'

c-----------------------------------------------------------------------
      subroutine control(ictrl,iolst)
c-----------------------------------------------------------------------
c
c --- SURFGEN   Version: 7.0.0          Level: 150330           CONTROL
c               D. Strimaitis, Exponent Inc.
c
c --- PURPOSE:  Obtain program control information
c               Open list file and write header
c
c --- INPUT:
c               ICTRL - integer   - Unit number of control file
c               IOLST - integer   - Unit number of output list file
c
c     Module CONST
c            cversion,clevel
c
c --- OUTPUT:
c     Module DATA (allocated arrays)
c            idone(),ispecial()
c            iceil(),icco(),icct(),iccs(),irh(),ipcode()
c            i8datim1(),i8datim2()
c            ws(),wd(),tk(),pres(),prate()
c     Module COUNTS (allocated arrays)
c            ncalmws(),nmissws(),nmisswd()
c            nmisstk(),nmissrh(),nmissp()
c            nmisscc(),nmissch()
c            
c     Module USERINF
c            (all)
c
c --- CONTROL called by: (main)
c --- CONTROL calls:     DELIM, MK_CASE, RM_BLANK, UTC_ZONE
c                        FMT_DATE, FMT_TIME, CAL_DAY, DATE_HRSEC
c                        COND_HR_DAY
c----------------------------------------------------------------------
c --- Control-file lines can be up to 320 characters long
c --- File pathnames can be up to 256 characters long
c----------------------------------------------------------------------
c --- Access module data
      use const
      use data
      use counts
      use userinf

      implicit none

c --- Declare arguments
      integer :: ictrl,iolst

c --- Locals
      logical :: lprocess, lend, lcsv, lbad
      integer :: i,nmet,ngroup,ieof, icase
      integer :: id1,id2,ieq
      integer :: l1,l2,r1,r2, num
      integer :: iotz,itz
      real    :: rotz,rtz
      character(len=320)  ::  blank320,aline320,bline320
      character(len=6)    ::  cminutes
      character(len=5)    ::  ulcase(2)
      character(len=1)    ::  dlim
      data dlim/'!'/
      data ulcase/'UPPER','LOWER'/

c --- Establish initial values (others in module userinf)
      do i=1,320
         blank320(i:i)=' '
      enddo

c -----------------------------------------------------------
c --- PASS 1
c -----------------------------------------------------------
c --- Scan control file for the list-file name, case flag, and
c --- number of input files (must allocate storage arrays)
c -----------------------------------------------------------
c --- (List file is not opened yet so error-writes will go to
c --- a FORT-file with number IOLST in name)
      nfiles=0
      ieof=0
      do while (ieof.EQ.0)
         aline320=blank320
         read(ictrl,'(a320)',iostat=ieof) aline320
         if(ieof.NE.0) EXIT
         call DELIM(dlim,aline320,id1,id2,ieq,lprocess)
         if(lprocess) then
c ---       Set range for variable name on left and value on right
            l1=id1+1
            l2=ieq-1
            r1=ieq+1
            r2=id2-1
            num=r2-r1+1
c ---       Convert to upper case
            call MK_CASE(iolst,'CONTROL','UPPER',aline320)
c ---       Interpret subset of inputs
            if(INDEX(aline320(l1:l2),'NFILES').GT.0) then
               read(aline320(r1:r2),'(i)') nfiles
            elseif(INDEX(aline320(l1:l2),'LSTFILE').GT.0) then
               if(num.GT.256) then
                  write(*,*)'Error in CONTROL: file pathname '//
     &                      'exceeds 256 characters' 
                  write(*,*) TRIM(aline320)
                  stop
               endif
               clstfile=blank320(1:256)
               clstfile=aline320(r1:r2)
c ---          Left-justify
               clstfile=ADJUSTL(clstfile)
            elseif(INDEX(aline320(l1:l2),'LCFILES').GT.0) then
               if(INDEX(aline320(r1:r2),'T').GT.0) then
                  LCFILES=.TRUE.
               else
                  LCFILES=.FALSE.
               endif
            endif
         endif
      enddo

c --- Set pointer for ULCASE (1=upper, 2=lower)
      icase=1
      if(lcfiles) icase=2

c --- Open list file and write header
      call MK_CASE(iolst,'CONTROL',ulcase(icase),clstfile)
      open(iolst,file=clstfile,status='unknown')
      write(iolst,*)
      write(iolst,*)
      write(iolst,*)'                    '//
     &              '   SURFGEN APPLICATION SUMMARY'
      write(iolst,*)'                   VERSION:'//cversion//
     &              '   LEVEL:'//clevel
      write(iolst,*)
      write(iolst,*)

c --- Rewind control file to complete the processing
      REWIND(ictrl)

c -------------------------------------------------------------
c --- Allocate arrays for storing input file information
      ALLOCATE (ctype(nfiles),cmetfile(nfiles))
      ALLOCATE (cmet6405(nfiles),cmet6406(nfiles))
      ALLOCATE (cstaid(nfiles),citzone(nfiles),cwsadj(nfiles))
      ALLOCATE (isectzadj(nfiles),isec2utc(nfiles))
      ALLOCATE (idone(nfiles),ispecial(nfiles))
      ALLOCATE (i8datim1(2,nfiles),i8datim2(2,nfiles))
      ALLOCATE (iceil(3,nfiles),icco(3,nfiles))
      ALLOCATE (icct(3,nfiles),iccs(3,nfiles))
      ALLOCATE (irh(3,nfiles),ipcode(3,nfiles))
      ALLOCATE (ws(3,nfiles),wd(3,nfiles),tk(3,nfiles))
      ALLOCATE (pres(3,nfiles),prate(3,nfiles))
c --- Allocate arrays for storing counts
      ALLOCATE (ncalm(nfiles))
      ALLOCATE (nmiss1(nvar,nfiles),nmiss(nvar,nfiles+1))
c -------------------------------------------------------------

c -----------------------------------------------------------
c --- PASS 2
c -----------------------------------------------------------
c --- Process control file again for all contents
c -----------------------------------------------------------
      ieof=0
      nmet=1
      ngroup=1
      lend=.FALSE.
      lcsv=.FALSE.
      lbad=.FALSE.

c --- Set character inputs to MISSING
      ctype='MISSING'
      cmetfile='MISSING'
      cmet6405='MISSING'
      cmet6406='MISSING'
      cstaid='MISSING'
      citzone='MISSING'
      cwsadj='MISSING'

c --- Loop over lines in control file
      do while (ieof.EQ.0)

c ---    Check if group terminator was just read on previous line
         if(lend) then
c ---       There are 3 input groups, and the third group contains
c ---       NFILES groups of repeated assignments: determine group
            if(ngroup.LT.3) then
c ---          Next group
               ngroup=ngroup+1
            else
c ---          Next met file in group 3
               nmet=nmet+1
            endif
c ---       Restore logical
            lend=.FALSE.
         endif

c ---    New line
         aline320=blank320
         read(ictrl,'(a320)',iostat=ieof) aline320
         if(ieof.NE.0) EXIT

c ---    Keep track of when the group terminator appears on a line
         bline320=aline320
         call RM_BLANK(bline320)
         call MK_CASE(iolst,'CONTROL','UPPER',bline320)
         if(INDEX(bline320,'!END!').GT.0) lend=.TRUE.

         call DELIM(dlim,aline320,id1,id2,ieq,lprocess)
         if(lprocess) then
c ---       Set range for variable name on left and value on right
            l1=id1+1
            l2=ieq-1
            r1=ieq+1
            r2=id2-1
            num=r2-r1+1
c ---       Convert to upper case
            call MK_CASE(iolst,'CONTROL','UPPER',aline320)
c ---       Check list of expected inputs

c ---       -----------------------------
c ---       Skip information obtained in first scan
            if(INDEX(aline320(l1:l2),'NFILES').GT.0) then
c ---          Nothing:  already processed
            elseif(INDEX(aline320(l1:l2),'LSTFILE').GT.0) then
c ---          Nothing:  already processed
            elseif(INDEX(aline320(l1:l2),'LCFILES').GT.0) then
c ---          Nothing:  already processed

c ---       -----------------------------
c ---       Next look for all filenames and retain internal blanks
            elseif(INDEX(aline320(l1:l2),'SRFFILE').GT.0) then
               if(num.GT.256) then
                  write(*,*)'Error in CONTROL: file pathname '//
     &                      'exceeds 256 characters' 
                  write(*,*) TRIM(aline320)
                  stop
               endif
               csrffile=blank320(1:256)
               csrffile=aline320(r1:r2)
               csrffile=ADJUSTL(csrffile)
               call MK_CASE(iolst,'CONTROL',ulcase(icase),csrffile)

            elseif(INDEX(aline320(l1:l2),'PRCFILE').GT.0) then
               if(num.GT.256) then
                  write(*,*)'Error in CONTROL: file pathname '//
     &                      'exceeds 256 characters' 
                  write(*,*) TRIM(aline320)
                  stop
               endif
               cprcfile=blank320(1:256)
               cprcfile=aline320(r1:r2)
               cprcfile=ADJUSTL(cprcfile)
               call MK_CASE(iolst,'CONTROL',ulcase(icase),cprcfile)

            elseif(INDEX(aline320(l1:l2),'CODEFILE').GT.0) then
               if(num.GT.256) then
                  write(*,*)'Error in CONTROL: file pathname '//
     &                      'exceeds 256 characters' 
                  write(*,*) TRIM(aline320)
                  stop
               endif
               cisd_code=blank320(1:256)
               cisd_code=aline320(r1:r2)
               cisd_code=ADJUSTL(cisd_code)
               call MK_CASE(iolst,'CONTROL',ulcase(icase),cisd_code)

            elseif(INDEX(aline320(l1:l2),'METFILE').GT.0) then
               if(num.GT.256) then
                  write(*,*)'Error in CONTROL: file pathname '//
     &                      'exceeds 256 characters' 
                  write(*,*) TRIM(aline320)
                  stop
               endif
               if(nmet.GT.nfiles) then
                  write(*,*)'Error in CONTROL: Number of met files'//
     &                      'exceeds specified NFILES = ',nfiles 
                  write(*,*) TRIM(aline320)
                  stop
               endif
               cmetfile(nmet)=blank320(1:256)
               cmetfile(nmet)=aline320(r1:r2)
               cmetfile(nmet)=ADJUSTL(cmetfile(nmet))
               call MK_CASE(iolst,'CONTROL',ulcase(icase),
     &                      cmetfile(nmet))

            elseif(INDEX(aline320(l1:l2),'MET6405').GT.0) then
               if(num.GT.256) then
                  write(*,*)'Error in CONTROL: file pathname '//
     &                      'exceeds 256 characters' 
                  write(*,*) TRIM(aline320)
                  stop
               endif
               if(nmet.GT.nfiles) then
                  write(*,*)'Error in CONTROL: Number of met files'//
     &                      'exceeds specified NFILES = ',nfiles 
                  write(*,*) TRIM(aline320)
                  stop
               endif
               cmet6405(nmet)=blank320(1:256)
               cmet6405(nmet)=aline320(r1:r2)
               cmet6405(nmet)=ADJUSTL(cmet6405(nmet))
               call MK_CASE(iolst,'CONTROL',ulcase(icase),
     &                      cmet6405(nmet))

            elseif(INDEX(aline320(l1:l2),'MET6406').GT.0) then
               if(num.GT.256) then
                  write(*,*)'Error in CONTROL: file pathname '//
     &                      'exceeds 256 characters' 
                  write(*,*) TRIM(aline320)
                  stop
               endif
               if(nmet.GT.nfiles) then
                  write(*,*)'Error in CONTROL: Number of met files'//
     &                      'exceeds specified NFILES = ',nfiles 
                  write(*,*) TRIM(aline320)
                  stop
               endif
               cmet6406(nmet)=blank320(1:256)
               cmet6406(nmet)=aline320(r1:r2)
               cmet6406(nmet)=ADJUSTL(cmet6406(nmet))
               call MK_CASE(iolst,'CONTROL',ulcase(icase),
     &                      cmet6406(nmet))

c ---       -----------------------------
c ---       All the remaining assignments do not have internal blanks
c ---       Remove blanks and reset pointers for rest
            else
               call RM_BLANK(aline320)
               call DELIM(dlim,aline320,id1,id2,ieq,lprocess)
c ---          Set range for variable name on left and value on right
               l1=id1+1
               l2=ieq-1
               r1=ieq+1
               r2=id2-1
               num=r2-r1+1

c ---          GROUP 1

c ---          [date and time may be either string entries]
               if(INDEX(aline320(l1:l2),'S_DATE').GT.0) then
                  cs_date=blank320(1:10)
                  cs_date(1:num)=aline320(r1:r2)
               elseif(INDEX(aline320(l1:l2),'S_TIME').GT.0) then
                  cs_time=blank320(1:8)
                  cs_time(1:num)=aline320(r1:r2)
               elseif(INDEX(aline320(l1:l2),'E_DATE').GT.0) then
                  ce_date=blank320(1:10)
                  ce_date(1:num)=aline320(r1:r2)
               elseif(INDEX(aline320(l1:l2),'E_TIME').GT.0) then
                  ce_time=blank320(1:8)
                  ce_time(1:num)=aline320(r1:r2)
c ---          [or individual integers]
               elseif(INDEX(aline320(l1:l2),'IBYR').GT.0) then
                  read(aline320(r1:r2),'(i)') kyrb
               elseif(INDEX(aline320(l1:l2),'IBMO').GT.0) then
                  read(aline320(r1:r2),'(i)') kmob
               elseif(INDEX(aline320(l1:l2),'IBDY').GT.0) then
                  read(aline320(r1:r2),'(i)') kdyb
               elseif(INDEX(aline320(l1:l2),'IBHR').GT.0) then
                  read(aline320(r1:r2),'(i)') khrb
               elseif(INDEX(aline320(l1:l2),'IBSEC').GT.0) then
                  read(aline320(r1:r2),'(i)') ksecb
               elseif(INDEX(aline320(l1:l2),'IEYR').GT.0) then
                  read(aline320(r1:r2),'(i)') kyre
               elseif(INDEX(aline320(l1:l2),'IEMO').GT.0) then
                  read(aline320(r1:r2),'(i)') kmoe
               elseif(INDEX(aline320(l1:l2),'IEDY').GT.0) then
                  read(aline320(r1:r2),'(i)') kdye
               elseif(INDEX(aline320(l1:l2),'IEHR').GT.0) then
                  read(aline320(r1:r2),'(i)') khre
               elseif(INDEX(aline320(l1:l2),'IESEC').GT.0) then
                  read(aline320(r1:r2),'(i)') ksece

               elseif(INDEX(aline320(l1:l2),'TIMESTEP').GT.0) then
                  read(aline320(r1:r2),'(i)') itimestep
               elseif(INDEX(aline320(l1:l2),'NSECDT').GT.0) then
                  read(aline320(r1:r2),'(i)') itimestep

               elseif(INDEX(aline320(l1:l2),'TSELECT').GT.0) then
                  ctselect=blank320(1:10)
                  ctselect(1:num)=aline320(r1:r2)

               elseif(INDEX(aline320(l1:l2),'OUTZONE').GT.0) then
                  cotzone=blank320(1:8)
                  cotzone(1:num)=aline320(r1:r2)
               elseif(INDEX(aline320(l1:l2),'ABTZ').GT.0) then
                  cotzone=blank320(1:8)
                  cotzone(1:num)=aline320(r1:r2)

               elseif(INDEX(aline320(l1:l2),'LSURF').GT.0) then
                  if(INDEX(aline320(r1:r2),'T').GT.0) then
                     LSURF=.TRUE.
                  else
                     LSURF=.FALSE.
                  endif
               elseif(INDEX(aline320(l1:l2),'LPRECIP').GT.0) then
                  if(INDEX(aline320(r1:r2),'T').GT.0) then
                     LPRECIP=.TRUE.
                  else
                     LPRECIP=.FALSE.
                  endif
               elseif(INDEX(aline320(l1:l2),'LPMISS').GT.0) then
                  if(INDEX(aline320(r1:r2),'T').GT.0) then
                     LPMISS=.TRUE.
                  else
                     LPMISS=.FALSE.
                  endif

c ---          GROUP 2

               elseif(INDEX(aline320(l1:l2),'NHEADR').GT.0) then
                  read(aline320(r1:r2),'(i)') nheadr
               elseif(INDEX(aline320(l1:l2),'DATE_COL').GT.0) then
                  read(aline320(r1:r2),'(i)') idate_col
               elseif(INDEX(aline320(l1:l2),'TIME_COL').GT.0) then
                  read(aline320(r1:r2),'(i)') itime_col
               elseif(INDEX(aline320(l1:l2),'WS_COL').GT.0) then
                  read(aline320(r1:r2),'(i)') iws_col
               elseif(INDEX(aline320(l1:l2),'WD_COL').GT.0) then
                  read(aline320(r1:r2),'(i)') iwd_col
               elseif(INDEX(aline320(l1:l2),'T_COL').GT.0) then
                  read(aline320(r1:r2),'(i)') it_col
               elseif(INDEX(aline320(l1:l2),'RH_COL').GT.0) then
                  read(aline320(r1:r2),'(i)') irh_col
               elseif(INDEX(aline320(l1:l2),'CLDHT_COL').GT.0) then
                  read(aline320(r1:r2),'(i)') icldht_col
               elseif(INDEX(aline320(l1:l2),'CEIL_COL').GT.0) then
                  read(aline320(r1:r2),'(i)') iceil_col
               elseif(INDEX(aline320(l1:l2),'CCVR_COL').GT.0) then
                  read(aline320(r1:r2),'(i)') iccvr_col
               elseif(INDEX(aline320(l1:l2),'P_COL').GT.0) then
                  read(aline320(r1:r2),'(i)') ip_col
               elseif(INDEX(aline320(l1:l2),'PR_COL').GT.0) then
                  read(aline320(r1:r2),'(i)') ipr_col

               elseif(INDEX(aline320(l1:l2),'DATE_FMT').GT.0) then
                  cdate_fmt=blank320(1:10)
                  cdate_fmt(1:num)=aline320(r1:r2)
               elseif(INDEX(aline320(l1:l2),'WS_UNIT').GT.0) then
                  cws_unit=blank320(1:10)
                  cws_unit(1:num)=aline320(r1:r2)
               elseif(INDEX(aline320(l1:l2),'WD_UNIT').GT.0) then
                  cwd_unit=blank320(1:10)
                  cwd_unit(1:num)=aline320(r1:r2)
               elseif(INDEX(aline320(l1:l2),'T_UNIT').GT.0) then
                  ct_unit=blank320(1:10)
                  ct_unit(1:num)=aline320(r1:r2)
               elseif(INDEX(aline320(l1:l2),'RH_UNIT').GT.0) then
                  crh_unit=blank320(1:10)
                  crh_unit(1:num)=aline320(r1:r2)
               elseif(INDEX(aline320(l1:l2),'CEIL_UNIT').GT.0) then
                  cceil_unit=blank320(1:10)
                  cceil_unit(1:num)=aline320(r1:r2)
               elseif(INDEX(aline320(l1:l2),'CCVR_UNIT').GT.0) then
                  cccvr_unit=blank320(1:10)
                  cccvr_unit(1:num)=aline320(r1:r2)
               elseif(INDEX(aline320(l1:l2),'P_UNIT').GT.0) then
                  cp_unit=blank320(1:10)
                  cp_unit(1:num)=aline320(r1:r2)
               elseif(INDEX(aline320(l1:l2),'PR_UNIT').GT.0) then
                  cpr_unit=blank320(1:10)
                  cpr_unit(1:num)=aline320(r1:r2)

c ---          GROUP 3  (file names already treated)

               elseif(INDEX(aline320(l1:l2),'FILETYPE').GT.0) then
                  ctype(nmet)=blank320(1:8)
                  ctype(nmet)(1:num)=aline320(r1:r2)
                  if(TRIM(ctype(nmet)).EQ.'CSV') lcsv=.TRUE.
               elseif(INDEX(aline320(l1:l2),'STAID').GT.0) then
                  cstaid(nmet)=blank320(1:20)
                  cstaid(nmet)(1:num)=aline320(r1:r2)
               elseif(INDEX(aline320(l1:l2),'WSADJ').GT.0) then
                  cwsadj(nmet)=blank320(1:8)
                  cwsadj(nmet)(1:num)=aline320(r1:r2)
               elseif(INDEX(aline320(l1:l2),'INZONE').GT.0) then
                  citzone(nmet)=blank320(1:8)
                  citzone(nmet)(1:num)=aline320(r1:r2)


               else
                  write(iolst,*)'Error in SURFGEN control file'
                  write(iolst,*)'Assignment variable unknown:'
                  write(iolst,*)TRIM(aline320)
                  stop 'HALTED in CONTROL -- see list file'
               endif
            endif
         endif
      enddo

c --- Condition date-time inputs and make associated variables
c --- Starting date-time
      if(cs_date .EQ. '9999-99-99') then
         call FMTDATE(iolst,'CONTROL','GROUP',cs_date,kyrb,kmob,kdyb)
      else
         call FMTDATE(iolst,'CONTROL','SPLIT',cs_date,kyrb,kmob,kdyb)
      endif
      if(cs_time .EQ. '99:99:99') then
         call FMTTIME(iolst,'CONTROL','GROUP',cs_time,khrb,ksecb)
      else
         call FMTTIME(iolst,'CONTROL','SPLIT',cs_time,khrb,ksecb)
      endif
      call CAL_DAY(iolst,'CONTROL','TOJULIAN',kyrb,kmob,kdyb,julb)
      call DATE_HRSEC(iolst,'CONTROL','TOSNGL',kyrb,julb,khrb,ksecb,
     &                k8bdatim)
      call COND_HR_DAY(iolst,'CONTROL','HR00S0000',k8bdatim)
      call DATE_HRSEC(iolst,'CONTROL','TOINDV',kyrb,julb,khrb,ksecb,
     &                k8bdatim)
      call CAL_DAY(iolst,'CONTROL','TOGREGOR',kyrb,kmob,kdyb,julb)
c --- Ending date-time
      if(ce_date .EQ. '9999-99-99') then
         call FMTDATE(iolst,'CONTROL','GROUP',ce_date,kyre,kmoe,kdye)
      else
         call FMTDATE(iolst,'CONTROL','SPLIT',ce_date,kyre,kmoe,kdye)
      endif
      if(ce_time .EQ. '99:99:99') then
         call FMTTIME(iolst,'CONTROL','GROUP',ce_time,khre,ksece)
      else
         call FMTTIME(iolst,'CONTROL','SPLIT',ce_time,khre,ksece)
      endif
      call CAL_DAY(iolst,'CONTROL','TOJULIAN',kyre,kmoe,kdye,jule)
      call DATE_HRSEC(iolst,'CONTROL','TOSNGL',kyre,jule,khre,ksece,
     &                k8edatim)
      call COND_HR_DAY(iolst,'CONTROL','HR00S0000',k8edatim)
      call DATE_HRSEC(iolst,'CONTROL','TOINDV',kyre,jule,khre,ksece,
     &                k8edatim)
      call CAL_DAY(iolst,'CONTROL','TOGREGOR',kyre,kmoe,kdye,jule)

c --- Time zone adjustment:  SECONDS to ADD to file times to get
c --- MODELING time zone
      call UTC_ZONE(iolst,'CONTROL','CHR2REAL',cotzone,rotz,iotz)
      do i=1,nfiles
         call UTC_ZONE(iolst,'CONTROL','CHR2REAL',citzone(i),rtz,itz)
         isectzadj(i)=NINT((rtz-rotz)*3600.)
         isec2utc(i)=NINT(rtz*3600.)
      enddo

c --- Check for wind adjustment configuration
      do i=1,nfiles
         if(cwsadj(i).NE.'ASOSBIAS' .AND. cwsadj(i).NE.'NONE    ') then
            lbad=.TRUE.
            write(iolst,*)
            write(iolst,*)'ERROR in Control File for Met File ',i
            write(iolst,*)'Wind Speed Adjustment: ',cwsadj(i)//
     &                    ' *** UNKNOWN (invalid) ***'
         endif
      enddo

c --- QA on file formats accepted
      do i=1,nfiles
         if(ctype(i).NE.'ISD     ' .AND.
     &      ctype(i).NE.'ASOS1   ' .AND.
     &      ctype(i).NE.'CSV     ') then
            lbad=.TRUE.
            write(iolst,*)
            write(iolst,*)'ERROR in Control File for Met File ',i
            write(iolst,*)'File Format Type: ',ctype(i)//
     &                    ' *** UNKNOWN (invalid) ***'
         endif
      enddo
c --- Temporary
      do i=1,nfiles
         if(ctype(i).EQ.'CSV     ') then
            lbad=.TRUE.
            write(iolst,*)
            write(iolst,*)'ERROR in Control File for Met File ',i
            write(iolst,*)'CSV format is not implemented'
         endif
      enddo

c --- Echo selected information to list file
      write(iolst,*)
      write(iolst,*)' Output Start: '//cs_date//'   '//cs_time
      write(iolst,*)'        End  : '//ce_date//'   '//ce_time
      write(iolst,*)'    Time Zone: '//cotzone
      write(iolst,*)'         Step: ',itimestep,'  (seconds)'
      if(ctselect.EQ.'LAST      ') then
         write(iolst,*)'Time Selected: ',ctselect//
     &                 ' valid in timestep'
      elseif(ctselect.EQ.'EXACT     ') then
         write(iolst,*)'Time Selected: ',ctselect//
     &                 ' match to end of timestep'
      else
         lbad=.TRUE.
         write(iolst,*)'Time Selected: ',ctselect//
     &                 ' *** UNKNOWN (invalid) ***'
      endif
      write(iolst,*)
      write(iolst,*)'ISD Codes read from File: '//TRIM(cisd_code)
      write(iolst,*)
      write(iolst,*)'Files Produced: '
      if(lsurf) write(iolst,*)' SURF.DAT   file: '//TRIM(csrffile)
      if(lsurf) write(iolst,*)' PRECIP.DAT file: '//TRIM(cprcfile)

      write(iolst,*)
      write(iolst,*)'Files Processed: '

      write(iolst,*)'                             Adjust '//
     &              '  Convert To '//cotzone
      write(iolst,*)'                             Wind   '//
     &              '               Add'
      write(iolst,*)'Type      ID                 Speed  '//
     &              '            (Minutes) Name'
      do i=1,nfiles
         cminutes='      '
         write(cminutes,'(i6)') (isectzadj(i)/60)
         if(ctype(i).EQ.'ASOS1   ') then
            write(iolst,*) ctype(i)//cstaid(i)//' '//cwsadj(i)//' '//
     &                     citzone(i)//'   '//cminutes//'   '//
     &                     TRIM(cmet6405(i))//' , '//
     &                     TRIM(cmet6406(i))
         else
            write(iolst,*) ctype(i)//cstaid(i)//' '//cwsadj(i)//' '//
     &                     citzone(i)//'   '//cminutes//'   '//
     &                     TRIM(cmetfile(i))
         endif
      enddo
      write(iolst,*)

      if(lcsv) then
         write(iolst,*)'CSV file attributes ---'
         write(iolst,*)'Number of header lines: ',nheadr
         write(iolst,*)'Date  column: ',idate_col,' ',cdate_fmt
         write(iolst,*)'Time  column: ',itime_col,' hh:mm'
         write(iolst,*)'WS    column: ',iws_col,' ',cws_unit
         write(iolst,*)'WD    column: ',iwd_col,' ',cwd_unit
         write(iolst,*)'T     column: ',it_col,' ',ct_unit
         write(iolst,*)'RH    column: ',irh_col,' ',crh_unit
         write(iolst,*)'P     column: ',ip_col,' ',cp_unit
         write(iolst,*)'PR    column: ',ipr_col,' ',cpr_unit
         write(iolst,*)'CLDHT column: ',icldht_col,' (no units)'
         write(iolst,*)'CEIL  column: ',iceil_col,' ',cceil_unit
         write(iolst,*)'CCVR  column: ',iccvr_col,' ',cccvr_unit
      endif

      if(lbad) stop 'HALT in CONTROL -- see list file'

      return
      end


c-----------------------------------------------------------------------
      subroutine setcodes(ioinp,iolst)
c-----------------------------------------------------------------------
c
c --- SURFGEN   Version: 7.0.0          Level: 150330          SETCODES
c               D. Strimaitis, Exponent Inc.
c
c --- PURPOSE:  Obtain coded information for ISD files from
c               ISD_CODES.DAT file
c
c --- INPUT:
c               IOINP - integer   - Unit number of ISD codes file
c               IOLST - integer   - Unit number of output list file
c
c --- OUTPUT:
c     Module ISDCODES
c            (All)
c
c --- SETCODES called by: (main)
c --- SETCODES calls:     MK_CASE, GETCOLS
c----------------------------------------------------------------------
c --- Lines in file can be up to 320 characters long
c----------------------------------------------------------------------
c --- Access module data
      use isdcodes

      implicit none

c --- Declare arguments
      integer :: ioinp,iolst

c --- Locals 
      logical ldb
      integer :: mxcol, ncol
      integer :: i,n,ieof,ncomm
      character(len=320)  ::  blank320,aline320
      character(len=80)   ::  acol(4)
      character(len=1)    ::  dlim, c1

      data dlim/':'/
      data mxcol/4/
      data ldb/.FALSE./

c --- Establish initial values (others in module userinf)
      do i=1,320
         blank320(i:i)=' '
      enddo

c --- PASS 1
c ----------

c --- Read all lines in file to count the number of ADD elements
      nadd_isd=0
      ieof=0
c --- First record
      aline320=blank320
      read(ioinp,'(a320)') aline320
      if(INDEX(aline320,'ISD_CODES.DAT').EQ.0) then
         write(iolst,*)'Error in SETCODES:  invalid codes file '
         write(iolst,*)'Expected ISD_CODES.DAT on first line'
         write(iolst,*)'Found: '//TRIM(aline320)
         stop 'HALT in SETCODES -- see list file'
      endif
c --- Skip comments
      read(ioinp,*) ncomm
      do i=1,ncomm
         read(ioinp,*)
      enddo
c --- Skip records until ADD section is reached
      i=0
      do while (i.EQ.0)
         aline320=blank320
         read(ioinp,'(a320)') aline320
         i=INDEX(aline320,'(ADD)')
      enddo
c --- Count number of ADD elements
      do while (ieof.EQ.0)
         c1=' '
         read(ioinp,'(a1)',iostat=ieof) c1
         if(ieof.NE.0) EXIT
         if(c1.NE.'>') nadd_isd=nadd_isd+1
      enddo

c --- Rewind file and allocate storage arrays
      REWIND(ioinp)
      ALLOCATE (code_add(nadd_isd),cdoc_add(nadd_isd))
      ALLOCATE (length_add(nadd_isd),iuse_add(nadd_isd))

c --- PASS 2
c ----------

      ieof=0
      n=0
      iuse_add=0
c --- Skip header & comments
      do i=1,ncomm+2
         read(ioinp,*)
      enddo

c --- Process all lines in file
      do while (ieof.EQ.0)
         aline320=blank320
         read(ioinp,'(a320)',iostat=ieof) aline320
         if(ieof.NE.0) EXIT
         if(aline320(1:1).NE.'>') then
c ---    Valid record
c ---       Break out columns into strings
            do i=1,4
               acol(i)=blank320(1:80)
            enddo
            call GETCOLS(dlim,aline320,mxcol,ncol,acol)
c ---       Expect 2 or 4 columns
            if(ncol.NE.2 .AND. ncol.NE.4) then
               write(iolst,*)'Error in SETCODES:  invalid #columns'
               write(iolst,*)'Expected 2 or 4 columns, Found ',ncol
               write(iolst,*)TRIM(aline320)
               stop 'HALT in SETCODES -- see list file'
            endif

c ---       Assign data
            if(ncol.EQ.2) then
               call MK_CASE(iolst,'SETCODES','UPPER',acol(1))
               if(INDEX(acol(1),'MAXLEN').GT.0) then
                  read(acol(2),'(i)') maxrecl_isd
               elseif(INDEX(acol(1),'CONTROL').GT.0) then
                  read(acol(2),'(i)') nctrl_isd
               elseif(INDEX(acol(1),'MANDATORY').GT.0) then
                  read(acol(2),'(i)') nmand_isd
               elseif(INDEX(acol(1),'Q_FLAGS').GT.0) then
                  cqfmand_isd=TRIM(acol(2))
               elseif(INDEX(acol(1),'Q_FAIL').GT.0) then
                  cfail_isd=TRIM(acol(2))
               else
                  write(iolst,*)'Error in SETCODES:  invalid record'
                  write(iolst,*)TRIM(aline320)
                  stop 'HALT in SETCODES -- see list file'
               endif
            else
               n=n+1
               code_add(n)=acol(1)(1:3)
               read(acol(2),'(i)') length_add(n)
               call MK_CASE(iolst,'SETCODES','UPPER',acol(3))
               if(INDEX(TRIM(acol(3)),'READ').GT.0) iuse_add(n)=1
               cdoc_add(n)=TRIM(acol(4))
            endif
         endif
      enddo     

      if(ldb) then
c ---    Echo results to list file
         write(iolst,*)
         write(iolst,*)'ISD Code Processing'
         write(iolst,*)'Max Record Length          = ',maxrecl_isd
         write(iolst,*)'Control Section Length     = ',nctrl_isd
         write(iolst,*)'Mandatory Section Length   = ',nmand_isd
         write(iolst,*)'Quality Flags in Mandatory = ',cqfmand_isd
         write(iolst,*)'Failure Flags in Mandatory = ',cfail_isd
         write(iolst,*)'List of ADDITIONAL Codes:'
         do n=1,nadd_isd
            write(iolst,'(a3,2i4,5x,a80)') code_add(n),length_add(n),
     &                                     iuse_add(n),cdoc_add(n)
         enddo
      endif

      return
      end

c-----------------------------------------------------------------------
      subroutine setup(iomet,iolst,iosurf,ioprec)
c-----------------------------------------------------------------------
c
c --- SURFGEN   Version: 7.0.0          Level: 150330             SETUP
c               D. Strimaitis, Exponent Inc.
c
c --- PURPOSE:  Prepare for processing
c               - Open all input met files
c               - QA expected contents using initial reads
c               - Open output file(s) and write header
c
c --- INPUT:
c               IOMET - integer   - Base-Unit number for all met files
c               IOLST - integer   - Unit number of output list file
c              IOSURF - integer   - Unit number of output SURF file
c              IOPREC - integer   - Unit number of output PRECIP file
c
c     Module CONST
c            cversion, clevel
c            izero, imiss4, rmiss4
c     Module HEAD
c            surfset,surfver,precset,precver,datadoc
c     Module USERINF
c            nfiles, cstaid(), cmetfile(), csrffile, cprcfile
c            cmet6405(),cmet6406()
c            kyrb,julb,khrb,ksecb,k8bdatim
c            kyre,jule,khre,ksece
c            cotzone
c
c --- OUTPUT:
c     Module DATA
c            n8datim1,n8datim2
c            nyr1,njul1,nhr1,nsec1,nyr2,njul2,nhr2,nsec2
c            idone()
c            i8datim1(),i8datim2()
c            ws(),wd(),tk(),pres(),prate()
c            iceil(),icco(),icct(),iccs(),irh(),ipcode()
c            ispecial()
c
c --- SETUP called by: (main)
c --- SETUP calls:     TIME_INCR, DATE_HRSEC, COND_HR_DAY
c                      READ_ISD, READ_ASOS1, MOVE_DAT, OPEN_ERR
c----------------------------------------------------------------------
c --- Access module data
      use const
      use data
      use head
      use userinf

      implicit none

c --- Declare arguments
      integer :: iomet,iolst,iosurf,ioprec

c --- Locals
      logical :: ldb
      integer :: i,io,io2,more,ierr
      integer :: iyr,imo,idy,jul,ihr,isec

      data ldb/.TRUE./

c --- Set the date-time for the first output period
c --- Start is directly from control file
      nyr1=kyrb
      njul1=julb
      nhr1=khrb
      nsec1=ksecb
      n8datim1=k8bdatim
c --- Increment one step for end date-time
      nyr2=kyrb
      njul2=julb
      nhr2=khrb
      nsec2=ksecb
      call TIME_INCR(iolst,'SETUP',nyr2,njul2,nhr2,nsec2,
     &               izero,itimestep)
      call DATE_HRSEC(iolst,'SETUP','TOSNGL',nyr2,njul2,nhr2,nsec2,
     &                n8datim2)

c --- Initialize data arrays to missing (time set to end)
      i8datim2=n8datim2
      i8datim1=n8datim2
      ws=rmiss4
      wd=rmiss4
      iceil=imiss4
      icco=imiss4
      icct=imiss4
      iccs=imiss4
      tk=rmiss4
      irh=imiss4
      pres=rmiss4
      ipcode=imiss4
      prate=rmiss4
      ispecial=imiss4

c --- Open input met files, process any headers, do first data reads
c --- The goal is to start COMP with the first available data that
c --- ends AFTER the first output period stored in (2,_), and the
c --- last available data before this stored in (1,_) and (3,_).
c --- But if no data are available in the first period, then missings
c --- are placed in (1,_) and (3,_).

      do i=1,nfiles
         io=iomet+i
         if(ctype(i).EQ.'ASOS1   ') then
            OPEN(io,file=TRIM(cmet6405(i)),form='formatted',
     &           status='old',iostat=ierr)
            if(ierr.NE.0) call OPEN_ERR(iolst,'SETUP','ASOS 6405 File',
     &                                  TRIM(cmet6405(i)),io)
            io2=10*iomet+i
            OPEN(io2,file=TRIM(cmet6406(i)),form='formatted',
     &           status='old',iostat=ierr)
            if(ierr.NE.0) call OPEN_ERR(iolst,'SETUP','ASOS 6406 File',
     &                                  TRIM(cmet6406(i)),io2)
         else
            OPEN(io,file=TRIM(cmetfile(i)),form='formatted',
     &           status='old',iostat=ierr)
            if(ierr.NE.0) call OPEN_ERR(iolst,'SETUP','Met Data File',
     &                                  TRIM(cmetfile(i)),io)
         endif
         idone(i)=0

         more=1
         do while (more.EQ.1)

            SELECT CASE (ctype(i))

            CASE ("ISD     ")
            call READ_ISD(io,iolst,i,
     &                    iyr,imo,idy,jul,ihr,isec,idone(i),
     &                    ws(2,i),wd(2,i),iceil(2,i),icco(2,i),
     &                    icct(2,i),iccs(2,i),tk(2,i),irh(2,i),
     &                    pres(2,i),ipcode(2,i),prate(2,i),
     &                    ispecial(i))
            if(idone(i).NE.0) then
               write(iolst,*)'ERROR in SETUP:  unexpected End-Of-File'
               write(iolst,*)'Processing File '//cmetfile(i)
               write(iolst,*)'Processing Period ',n8datim1,
     &                       ' to ',n8datim2
               stop 'HALT in SETUP -- see list file'
            endif

            CASE ("ASOS1   ")
            call READ_ASOS1(io,io2,iolst,i,
     &                    iyr,imo,idy,jul,ihr,isec,idone(i),
     &                    ws(2,i),wd(2,i),iceil(2,i),icco(2,i),
     &                    icct(2,i),iccs(2,i),tk(2,i),irh(2,i),
     &                    pres(2,i),ipcode(2,i),prate(2,i),
     &                    ispecial(i))
            if(idone(i).NE.0) then
               write(iolst,*)'ERROR in SETUP:  unexpected End-Of-File'
               write(iolst,*)'Processing File '//cmet6405(i)
               write(iolst,*)'            and '//cmet6406(i)
               write(iolst,*)'Processing Period ',n8datim1,
     &                       ' to ',n8datim2
               stop 'HALT in SETUP -- see list file'
            endif

            CASE ("CSV     ")

            END SELECT

c ---       Process time of this observation as end-time
            call DATE_HRSEC(iolst,'SETUP','TOSNGL',iyr,jul,ihr,isec,
     &                      i8datim2(2,i))
            call COND_HR_DAY(iolst,'SETUP','HR00S0000',i8datim2(2,i))
c ---       Set begin=end time like old SURF format
            i8datim1(2,i)=i8datim2(2,i)
c ---       Compare to date-time of output period
            if(i8datim2(2,i).LT.n8datim1) then
c ---          Record ends before start of first output period: missing
               more=1
               CYCLE
            elseif(i8datim2(2,i).LE.n8datim2) then
c ---          Record is in the output period
c ---          Move these data into (1,i) or (3,i)
c ---          Set the time
               call MOVE_DAT(iolst,'DATETIME',i,2,1)
               if(ispecial(i).EQ.0) then
c ---             Regular observation
                  call MOVE_DAT(iolst,'UPDATE',i,2,1)
               else
c ---             Special observation
                  call MOVE_DAT(iolst,'UPDATE',i,2,3)
               endif
               more=1
               CYCLE
            else
c ---          Record is after the output period
               more=0
               EXIT
            endif
         enddo

c ---    Debug
         if(LDB) then
         write(iolst,*)
         write(iolst,*)'Met data at startup, station index ',i
         write(iolst,*)'First output period: ',n8datim1,n8datim2
         write(iolst,*)'Time 1 Ends YYYYJJJHHSSSS: ',i8datim2(1,i)
         write(iolst,*)'Time 2 Ends YYYYJJJHHSSSS: ',i8datim2(2,i)
         write(iolst,*)'Variable     Time 1      Time 2    Special'
         write(iolst,*)'ws()    ',ws(1,i),ws(2,i),ws(3,i)
         write(iolst,*)'wd()    ',wd(1,i),wd(2,i),wd(3,i)
         write(iolst,*)'tk()    ',tk(1,i),tk(2,i),tk(3,i)
         write(iolst,*)'pres()  ',pres(1,i),pres(2,i),pres(3,i)
         write(iolst,*)'prate() ',prate(1,i),prate(2,i),prate(3,i)
         write(iolst,*)'iceil() ',iceil(1,i),iceil(2,i),iceil(3,i)
         write(iolst,*)'icco()  ',icco(1,i),icco(2,i),icco(3,i)
         write(iolst,*)'icct()  ',icct(1,i),icct(2,i),icct(3,i)
         write(iolst,*)'iccs()  ',iccs(1,i),iccs(2,i),iccs(3,i)
         write(iolst,*)'irh()   ',irh(1,i),irh(2,i),irh(3,i)
         write(iolst,*)'ipcode()',ipcode(1,i),ipcode(2,i),ipcode(3,i)
         write(iolst,*)'ispecial',ispecial(i)
         endif
      enddo

c --- Open output files and write header
      if(lsurf) then
         ncomment=nfiles+3
         OPEN(iosurf,file=TRIM(csrffile),form='formatted',
     &        status='unknown')
         write(iosurf,'(2a16,a64)') surfset,surfver,datadoc
         write(iosurf,'(i3)') ncomment
         write(iosurf,'(a)') 'Created by SURFGEN   Version: '//
     &                        TRIM(cversion)//' Level: '//TRIM(clevel)
         write(iosurf,'(a)') 'Wind Speed Adjustments (if any):'
         write(iosurf,'(a)')'Type      ID                 Adjustment'
         do i=1,nfiles
            write(iosurf,*) ctype(i)//cstaid(i)//' '//cwsadj(i)
         enddo
         write(iosurf,'(a8)') 'NONE    '
         write(iosurf,'(a8)') cotzone
         write(iosurf,'(2(i6,i4.3,i4.2,i6.4),i6)')
     &         kyrb,julb,khrb,ksecb,kyre,jule,khre,ksece,nfiles
         do i=1,nfiles
            write(iosurf,'(a20)') cstaid(i)
         enddo
      endif
      if(lprecip) then
         OPEN(ioprec,file=TRIM(cprcfile),form='formatted',
     &        status='unknown')
         write(ioprec,'(2a16,a64)') precset,precver,datadoc
         write(ioprec,'(a1)') '1'
         write(ioprec,'(a)') 'Created by SURFGEN   Version: '//
     &                        TRIM(cversion)//' Level: '//TRIM(clevel)
         write(ioprec,'(a8)') 'NONE    '
         write(ioprec,'(a8)') cotzone
         write(ioprec,'(2(i6,i4.3,i4.2,i6.4),i6)')
     &         kyrb,julb,khrb,ksecb,kyre,jule,khre,ksece,nfiles
         do i=1,nfiles
            write(ioprec,'(a20)') cstaid(i)
         enddo
      endif

      return
      end

c-----------------------------------------------------------------------
      subroutine comp(iomet,iolst,iosurf,ioprec)
c-----------------------------------------------------------------------
c
c --- SURFGEN   Version: 7.0.0          Level: 150330              COMP
c               D. Strimaitis, Exponent Inc.
c
c --- PURPOSE:  Loop over all station files and output data for
c               requested time period
c
c --- INPUT:
c               IOMET - integer   - Base-Unit number for all met files
c               IOLST - integer   - Unit number of output list file
c              IOSURF - integer   - Unit number of output SURF file
c              IOPREC - integer   - Unit number of output PRECIP file
c
c     Module CONST
c            imiss4, rmiss4
c     Module DATA
c            n8datim1,n8datim2
c            nyr1,njul1,nhr1,nsec1,nyr2,njul2,nhr2,nsec2
c            idone(),ispecial()
c            i8datim1(),i8datim2()
c            ws(),wd(),tk(),pres(),prate()
c            iceil(),icco(),icct(),iccs(),irh(),ipcode()
c     Module USERINF
c            ctype()
c            nfiles, itimestep, lpmiss
c            kyre,jule,khre,ksece,k8edatim
c
c --- OUTPUT:
c     Module COUNTS
c            npdproc
c            ncalm(),nmiss1(),nmiss()
c
c --- COMP called by: (main)
c --- COMP calls:     TIMECHK, TIME_INCR, DATE_HRSEC, COND_HR_DAY
c                     READ_ISD, READ_ASOS1, MOVE_DAT
c----------------------------------------------------------------------
c ---       Helpful Definitions:
c ---       n8datim1=start of current output period
c ---       n8datim2=end of current output period
c ---       i8datim2(1,_)=end-time of "current" observation (1,_)
c ---       i8datim2(2,_)=end-time of "next" observation (2,_)
c ---       var(1,_)= "current" regular observation
c ---       var(2,_)= "next" regular or special observation
c ---       var(3,_)= "current" special observation
c----------------------------------------------------------------------
c --- Access module data
      use const
      use data
      use counts
      use userinf

      implicit none

c --- Declare arguments
      integer :: iomet,iolst,iosurf,ioprec

c --- Locals
      logical :: lcloudlst
      integer :: i,k,io,io2, ires1,ires2, more
      integer :: iyr,imo,idy,jul,ihr,isec
      integer :: icc,nmissvar

c --- Zero missings-counter array for the full output period
      nmiss=0
c --- Zero calms-counter array for the full output period
      ncalm=0

c --- List-file output when all stations are missing cloud data are
c --- not helpfull when all stations are ASOS1 type since these
c --- files do not contain cloud data -- set a logical for this
      lcloudlst=.FALSE.
      do i=1,nfiles
         if(INDEX(ctype(i),'ASOS1').EQ.0) lcloudlst=.TRUE.
      enddo

c --- Loop over all times requested
c --- SETUP has already established time and data for first period
      do while (n8datim2.LE.k8edatim)
c ---    Write time record
c         if(lsurf) write(iosurf,'(2(i6,i4.3,i4.2,i6.4))')
         if(lsurf) write(iosurf,'(2(3i4,i5,3x))')
     &                   nyr1,njul1,nhr1,nsec1,
     &                   nyr2,njul2,nhr2,nsec2
         if(lprecip) write(ioprec,'(2(i6,i4.3,i4.2,i6.4))')
     &                     nyr1,njul1,nhr1,nsec1,
     &                     nyr2,njul2,nhr2,nsec2

c ---    Zero missings-counter array for this period
         nmiss1=0

c ---    Loop over station-files
         do i=1,nfiles
            io=iomet+i
            io2=10*iomet+i
            more=1
            do while (more.EQ.1)
c ---          Determine data for output period
c ---          Overlap of "current" observation and output period
               call TIMECHK(iolst,'OVERLAP',i8datim1(1,i),
     &                      i8datim2(1,i),n8datim1,n8datim2,ires1)
c ---          Overlap of "next" observation and output period
               call TIMECHK(iolst,'OVERLAP',i8datim1(2,i),
     &                      i8datim2(2,i),n8datim1,n8datim2,ires2)
               if(ires1.EQ.99) then
c ---             Current obs spans output period so use it
                  more=0
               elseif(ires2.EQ.99) then
c ---             Next obs spans output period so move to current
                  call MOVE_DAT(iolst,'DATETIME',i,2,1)
                  if(ispecial(i).EQ.0) then
c ---                Regular gets the data
                     call MOVE_DAT(iolst,'REPLACE',i,2,1)
c ---                Special set to missing
                     call MOVE_DAT(iolst,'MISSING',i,0,3)
                  else
c ---                Special gets the data
                     call MOVE_DAT(iolst,'REPLACE',i,2,3)
c ---                Regular set to missing
                     call MOVE_DAT(iolst,'MISSING',i,0,1)
                  endif
                  more=1
               elseif(ires1.EQ.-2) then
c ---             Current obs preceeds output period
                  if(ires2.EQ.2) then
c ---                Next obs begins/ends after output period
c ---                Output period is not in file
                     more=0
c ---                Reset current time to output time
                     i8datim1(1,i)=n8datim1
                     i8datim2(1,i)=n8datim2
c ---                Set all (1,_) and (3,_) data to missing
                     call MOVE_DAT(iolst,'MISSING',i,0,1)
                     call MOVE_DAT(iolst,'MISSING',i,0,3)
                  else
c ---                Shift next obs into current and read next
                     call MOVE_DAT(iolst,'DATETIME',i,2,1)
                     if(ispecial(i).EQ.0) then
c ---                   Regular gets the data
                        call MOVE_DAT(iolst,'REPLACE',i,2,1)
c ---                   Special set to missing
                        call MOVE_DAT(iolst,'MISSING',i,0,3)
                     else
c ---                   Special gets the data
                        call MOVE_DAT(iolst,'REPLACE',i,2,3)
c ---                   Regular set to missing
                        call MOVE_DAT(iolst,'MISSING',i,0,1)
                     endif
                     more=1
                  endif
               elseif(ires1.EQ.-1. OR. ires1.EQ.0) then
c ---             Current obs ends in output period
                  if(ires2.EQ.1 .OR. ires2.EQ.2) then
c ---                Next obs ends after output period (do not use yet)
                     more=0
                  else
c ---                Next also ends in output period (preferred)
c ---                Update current and read next
                     call MOVE_DAT(iolst,'DATETIME',i,2,1)
                     if(ispecial(i).EQ.0) then
c ---                   Regular Obs
                        call MOVE_DAT(iolst,'UPDATE',i,2,1)
                     else
c ---                   Special Obs
                        call MOVE_DAT(iolst,'UPDATE',i,2,3)
                     endif
                     more=1
                  endif
               endif
               if(more.EQ.1) then
c ---             Need to read a new "next" data record

c ---             Read data from met file by type
c---              -------------------------------
                  SELECT CASE (ctype(i))

                  CASE ("ISD     ")
                     call READ_ISD(io,iolst,i,iyr,imo,idy,jul,ihr,isec,
     &                             idone(i),ws(2,i),wd(2,i),iceil(2,i),
     &                             icco(2,i),icct(2,i),iccs(2,i),
     &                             tk(2,i),irh(2,i),pres(2,i),
     &                             ipcode(2,i),prate(2,i),ispecial(i))
                     if(idone(i).NE.0) then
c ---                   No more records in this file
c ---                   Set to missing
                        call MOVE_DAT(iolst,'MISSING',i,0,2)
                        ispecial(i)=0
c ---                   Set time to end of output period + 24 hours
                        iyr=kyre
                        jul=jule
                        ihr=khre
                        isec=ksece
                        call TIME_INCR(iolst,'COMP',iyr,jul,ihr,isec,
     &                                 24,0)
                     endif
c ---                Process time of this observation as end-time
                     call DATE_HRSEC(iolst,'SETUP','TOSNGL',
     &                               iyr,jul,ihr,isec,i8datim2(2,i))
                     call COND_HR_DAY(iolst,'SETUP','HR00S0000',
     &                                i8datim2(2,i))
c ---                Set begin=end time like old SURF format
                     i8datim1(2,i)=i8datim2(2,i)

                  CASE ("ASOS1   ")
                   call READ_ASOS1(io,io2,iolst,i,
     &                             iyr,imo,idy,jul,ihr,isec,
     &                             idone(i),ws(2,i),wd(2,i),iceil(2,i),
     &                             icco(2,i),icct(2,i),iccs(2,i),
     &                             tk(2,i),irh(2,i),pres(2,i),
     &                             ipcode(2,i),prate(2,i),ispecial(i))
                     if(idone(i).NE.0) then
c ---                   No more records in this file
c ---                   Set to missing
                        call MOVE_DAT(iolst,'MISSING',i,0,2)
                        ispecial(i)=0
c ---                   Set time to end of output period + 24 hours
                        iyr=kyre
                        jul=jule
                        ihr=khre
                        isec=ksece
                        call TIME_INCR(iolst,'COMP',iyr,jul,ihr,isec,
     &                                 24,0)
                     endif
c ---                Process time of this observation as end-time
                     call DATE_HRSEC(iolst,'SETUP','TOSNGL',
     &                               iyr,jul,ihr,isec,i8datim2(2,i))
                     call COND_HR_DAY(iolst,'SETUP','HR00S0000',
     &                                i8datim2(2,i))
c ---                Set begin=end time like old SURF format
                     i8datim1(2,i)=i8datim2(2,i)

                  CASE ("CSV     ")

                  END SELECT
               endif
            enddo

c ---       Combine special(3,_) and regular(1,_) values as needed to
c ---       fill missings in regular observations
            call MOVE_DAT(iolst,'FILLMISS',i,3,1)

c ---       Apply LPMISS directive for precip
            if(lpmiss) then
c ---          Report zero precip when rate/code is missing in valid obs
               if(prate(1,i).GT.(rmiss4-1.0) .AND.
     &            ipcode(1,i).GT.(imiss4-1)) then
c ---             Expect a valid observation to include at least one
c ---             of the following:
                  if(ws(1,i).LT.(rmiss4-1.0) .OR.
     &               wd(1,i).LT.(rmiss4-1.0) .OR.
     &               tk(1,i).LT.(rmiss4-1.0) .OR.
     &               irh(1,i).LT.(imiss4-1)) then
                     prate(1,i)=0.0
                     ipcode(1,i)=0
                  endif
               endif
            endif

c ---       Select preferred cloud-cover for output
            icc=icco(1,i)
            if(icc.EQ.imiss4) icc=icct(1,i)
            if(icc.EQ.imiss4) icc=iccs(1,i)

c ---       Write output record(s) for this station-file
            if(lsurf) write(iosurf,'(2f9.3,2i5,2(f9.3,i5))')
     &                      ws(1,i),wd(1,i),iceil(1,i),icc,
     &                      tk(1,i),irh(1,i),pres(1,i),ipcode(1,i)
            if(lprecip) write(ioprec,'(f9.3)') prate(1,i)


c ---       Update counts of missing surface data at each station
            if(ws(1,i).GT.(rmiss4-1.))    nmiss1(1,i)=1
            if(wd(1,i).GT.(rmiss4-1.))    nmiss1(2,i)=1
            if(tk(1,i).GT.(rmiss4-1.))    nmiss1(3,i)=1
            if(irh(1,i).GT.(imiss4-1))    nmiss1(4,i)=1
            if(pres(1,i).GT.(rmiss4-1.))  nmiss1(5,i)=1
            if(iceil(1,i).GT.(imiss4-1))  nmiss1(6,i)=1
            if(icc.GT.(imiss4-1))         nmiss1(7,i)=1
            if(ipcode(1,i).GT.(imiss4-1)) nmiss1(8,i)=1

c ---       Update counts of calms at each station
            if(ws(1,i).LT.0.01) ncalm(i)=ncalm(i)+1

         enddo

c ---    Update counters for this period
c ---    Assess each output variable
         do k=1,nvar
            nmissvar=0
c ---       Collect information from each station
            do i=1,nfiles
               nmiss(k,i)=nmiss(k,i)+nmiss1(k,i)
               if(nmiss1(k,i).EQ.1) nmissvar=nmissvar+1
            enddo
            if(nmissvar.EQ.nfiles) then
c ---          This variable is missing at all stations
               nmiss(k,nfiles+1)=nmiss(k,nfiles+1)+1
c ---          Report this period to the list file, except for cloud
c ---          cover and ceiling height if all stations are ASOS1
               if(lcloudlst .OR. (k.NE.6 .AND. k.NE.7)) 
     &            write(iolst,*)'All stations missing '//cvar(k)//
     &                          'at period ending (YYYYJJJHHSSSS) ',
     &                          n8datim2,'  '//cotzone
            endif
         enddo

c ---    Update periods processed
         npdproc=npdproc+1

c ---    Update output period
         nyr1=nyr2
         njul1=njul2
         nhr1=nhr2 
         nsec1=nsec2
         n8datim1=n8datim2
         call TIME_INCR(iolst,'COMP',nyr2,njul2,nhr2,nsec2,
     &                  0,itimestep)
         call DATE_HRSEC(iolst,'SETUP','TOSNGL',
     &                   nyr2,njul2,nhr2,nsec2,n8datim2)

      enddo

      return
      end

c-----------------------------------------------------------------------
      subroutine read_isd(io,iolst,k,iyr,imo,idy,jul,ihr,isec,ieof,
     &                    ws,wd,iceil,icco,icct,iccs,tk,irh,pres,
     &                    ipcode,prate,iobs)
c-----------------------------------------------------------------------
c
c --- SURFGEN   Version: 7.0.0          Level: 150330          READ_ISD
c               D. Strimaitis, Exponent Inc.
c
c --- PURPOSE:  Extracts data from 1 record of ISD-format file
c
c --- INPUT:
c                  IO - integer   - File unit number for ISD file
c               IOLST - integer   - Unit number of list file
c                   K - integer   - Current station (file) index
c
c     Module CONST
c            izero, imiss4, rmiss4
c     Module ISDCODES
c            maxrecl_isd
c     Module USERINF
c            cstaid(),cwsadj(),isectzadj()
c
c --- OUTPUT:
c
c ---          Date-Time is END of observation period
c                 IYR - integer   - Year (YYYY)
c                 IMO - integer   - Month (MM)
c                 IDY - integer   - Day (DD)
c                 JUL - integer   - Julian Day (DDD)
c                 IHR - integer   - Hour (hh)
c                ISEC - integer   - Second (ssss)
c
c                IEOF - integer   - Completion flag (end of file)
c                  WS - real      - Wind speed (m/s)
c                  WD - real      - Wind direction
c               ICEIL - integer   - Ceiling height (100s feet)
c                ICCO - integer   - Opaque cloud cover
c                ICCT - integer   - Total cloud cover
c                ICCS - integer   - Summation cloud cover
c                  TK - real      - Air temperature (K)
c                 IRH - integer   - Relative humidity (%)
c                PRES - real      - Station pressure  (mb)
c              IPCODE - integer   - Precipitation code
c                                      0 = No precip
c                                      1 = Liquid precip
c                                     20 = Frozen precip
c                                   9999 = Missing
c               PRATE - real      - Liquid precip rate (mm/hr)
c                IOBS - integer   - Observation flag
c                                   -1: drop observation
c                                    0: regular observation
c                                    1: special observation
c
c --- READ_ISD called by: SETUP, COMP
c --- READ_ISD calls:     CAL_DAY, TIME_INCR, MK_CASE
c----------------------------------------------------------------------
c --- Notes
c 1.  Observations stamped at minute KMIN or later in the output hour
c     are processed, where KMIN is the integer minute of the time.
c     Times at the END of the hour have minute=00, and are included.
c 2.  Observations are flagged from report type for later interpretation
c     (e.g., regular and special observations)
c 3.  Calling program determines if values returned overwrite previous
c     values found for the period
c 4.  Sky Cloud Cover summation involves multiple observations on one
c     record:
c        - always use the last (deepest layer) summation code GDn
c        - when code GDn is missing, use maximum of the layer code GAn
c----------------------------------------------------------------------
c --- Access modules
      use const
      use isdcodes
      use userinf

      implicit none

c --- Declare arguments
      integer        :: io,iolst,k
      integer        :: iyr,imo,idy,jul,ihr,isec
      integer        :: ieof,iceil,icco,icct,iccs,irh,ipcode,iobs
      real           :: ws,wd,tk,pres,prate

c --- Locals
      logical        :: ldb
      character*3200 :: aline
      character*20   :: aid20,awmo20,ausaf20
      character*6    :: aobs,adrop(5),aspecial(3)
      character*3    :: aqceil3,char3
      character*1    :: aqws,aqwd,aqwt,aqtk,aqdp,aqslp,chk,aqflag
      character*1    :: aqflg1,aqflg2
      integer        :: i,n,nlast,imatch, imm
      integer        :: iwsi,iwdi,itki,idpi,iceili,ipri
      integer        :: i1,i2,nhr
      integer        :: icci,icc2i,icca,iccb,icldmap7(7),icldmap11(11)
      integer        :: islpi,izi
      integer        :: ipstn,ipalt
      integer        :: ipca(100),ipcm(100),ipctk,ipc,ipw
      integer        :: ndrop,nspecial,kmin,iwskt
      real           :: slp,zstn,dpf,tf
      real           :: fkt2mps,add_kt

c --- ASOS bias adjustment
      data fkt2mps/0.5144/, add_kt/0.5/

c --- Cloud cover mapping to tenths
      data icldmap7/0,2,4,8,10,10,10/
      data icldmap11/0,1,3,4,5,6,8,9,10,10,10/

c --- Observation types to drop
      data ndrop/5/,adrop/'BOGUS ','COOPD ','PCP15 ','SOD   ','SOM   '/
c --- Special observations
      data nspecial/3/,aspecial/'FM-12 ','FM-16 ','SOASP '/

c --- Drop times BEFORE kmin minutes past the hour (but not 00)
      data kmin/1/

c --- Precip code mapping to present weather (00 to 99)
c ---   -1 = Not observed or indeterminate
c ---    0 = No precip
c ---    1 = Liquid precip
c ---   20 = Frozen precip
c --- Manual
      data ipcm/ 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
     &           0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
     &           1, 1,20, 1, 1, 1,-1,-1,-1,-1,
     &           0, 0, 0, 0, 0, 0,20,20,20,20,
     &           0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
     &           1, 1, 1, 1, 1, 1, 1, 1, 1, 1,
     &           1, 1, 1, 1, 1, 1, 1, 1, 1, 1,
     &          20,20,20,20,20,20,20,20,20,20,
     &           1, 1, 1, 1, 1,20,20,-1,-1,-1,
     &          -1, 1, 1,-1,-1,-1,20,-1,-1,20/
c --- Automatic
      data ipca/ 0, 0, 0, 0, 0, 0,-1,-1,-1,-1,
     &           0,20, 0,-1,-1,-1,-1,-1, 0,-1,
     &          -1, 1, 1, 1,20, 1,-1,20,20,20,
     &           0, 0, 0, 0, 0, 0,-1,-1,-1,-1,
     &           1, 1, 1, 1, 1,20,20, 1, 1,-1,
     &           1, 1, 1, 1, 1, 1, 1, 1, 1,-1,
     &           1, 1, 1, 1, 1, 1, 1, 1, 1,-1,
     &          20,20,20,20,20,20,20,20,20,-1,
     &           1, 1, 1, 1, 1,20,20,20,20,20,
     &          -1, 0,-1,20, 0,-1,20,-1,-1, 0/

c --- Make sure that line used to read data record exceeds stated max
      if(3200.LE.maxrecl_isd) then
         write(iolst,*)'ERROR in READ_ISD:  full record is too long'
         write(iolst,*)'Size in program = 3200'
         write(iolst,*)'Size in file    = ',maxrecl_isd
         stop 'HALT in READ_ISD -- see list file'
      endif

c --- Initialize all output variables to missing
      ws=rmiss4
      wd=rmiss4
      tk=rmiss4
      irh=imiss4
      pres=rmiss4
      iceil=imiss4
      icco=imiss4
      icct=imiss4
      iccs=imiss4
      ipcode=imiss4
      prate=rmiss4
c --- Selected locals
      slp=rmiss4
      zstn=rmiss4
      icca=imiss4
      iccb=imiss4
      ipalt=imiss4
      ipstn=imiss4
      ipc=imiss4

c --- Read 1 record from file
      ieof=0
      do while (ieof.EQ.0)
c ---    Rely on padding to initialize aline with blanks
         aline=' '
         read(io,'(a3200)',iostat=ieof) aline
         if(ieof.NE.0) return
c ---    Drop times that are too early in the hour
         read(aline(26:27),'(i2)') imm
         if(imm.NE.0 .AND. imm.LT.kmin) CYCLE
c ---    Assess observation type
         iobs=0
         do i=1,ndrop
            if(aline(42:46).EQ.adrop(i)) iobs=-1
         enddo
         do i=1,nspecial
            if(aline(42:46).EQ.aspecial(i)) iobs=1
         enddo
c ---    Got one?
         if(iobs.GE.0) EXIT
      enddo

c --- Check station ID (allow 20 characters - trailing blanks)
      aid20='                    '
      awmo20='                    '
      ausaf20='                    '
      aid20(1:11)=aline(5:15)
      ausaf20(1:6)=aline(5:10)
      awmo20(1:5)=aline(11:15)
      if(cstaid(k).NE.ausaf20 .AND. cstaid(k).NE.awmo20 .AND.
     &   cstaid(k).NE.aid20) then
         write(iolst,*)'ERROR in READ_ISD:  Station ID not found'
         write(iolst,*)'Requested ID = ',cstaid(k)
         write(iolst,*)'ID in file   = ',aid20
         write(iolst,*) TRIM(aline)
         stop 'HALT in READ_ISD -- see list file'
      endif

c --- Extract date-time and adjust time zone
      read(aline(16:19),'(i4)') iyr
      read(aline(20:21),'(i2)') imo
      read(aline(22:23),'(i2)') idy
      read(aline(24:25),'(i2)') ihr
      read(aline(26:27),'(i2)') imm
c --- Minutes to seconds
      isec=imm*60

c --- Adjust time to output time zone
      call CAL_DAY(iolst,'READ_ISD','TOJULIAN',iyr,imo,idy,jul)
      call TIME_INCR(iolst,'READ_ISD',iyr,jul,ihr,isec,izero,
     &               isectzadj(k))
      call CAL_DAY(iolst,'READ_ISD','TOGREGOR',iyr,imo,idy,jul)

c --- Get station elevation (redundant after first time ...)
      read(aline(47:51),'(i5)') izi
c --- In meters (OK)
      if(izi.LT.9999) zstn=FLOAT(izi)

c --- Temporary debug (screen by Date-Time LST)
      ldb=.FALSE.
c      if(jul.EQ.1 .AND. (ihr.GE.0 .AND. ihr.LE.6)) then
c         ldb=.TRUE.
c         write(iolst,*)
c         write(iolst,*)'STATION: ',cstaid(k)
c         write(iolst,*)'JUL,HR,MM = ',jul,ihr,imm,'  '//cotzone
c      endif


c --- MANDATORY Data Section
c ---------------------------

c --- Pull data from mandatory section of record
      read(aline(61:105),'(i3,2a1,i4,a1,i5,a3,9x,3(i5,a1))')
     &    iwdi,aqwd,aqwt,iwsi,aqws,iceili,aqceil3,
     &    itki,aqtk,idpi,aqdp,islpi,aqslp

c --- Screen for known quality flags
      chk=aqwd
      if(INDEX(cqfmand_isd,chk).EQ.0) goto 999
      chk=aqws
      if(INDEX(cqfmand_isd,chk).EQ.0) goto 999
      chk=aqceil3(1:1)
      if(INDEX(cqfmand_isd,chk).EQ.0) goto 999
      chk=aqtk
      if(INDEX(cqfmand_isd,chk).EQ.0) goto 999
      chk=aqdp
      if(INDEX(cqfmand_isd,chk).EQ.0) goto 999
      chk=aqslp
      if(INDEX(cqfmand_isd,chk).EQ.0) goto 999

c --- Use valid values

c --- Wind
c --- Force upper case on Wind Type (aqwt) codes
      call MK_CASE(iolst,'READ_ISD','UPPER',aqwt)
c --- Assess Calm and Variable Wind Type codes first
      if(aqwt.EQ.'C') then
c ---    Calm (regardless of numerical values)
         ws=0.0
         wd=0.0
      elseif(aqwt.EQ.'V') then
c ---    Variable = missing (regardless of numerical values)
c ---    Keep initial missing values
      elseif(INDEX(cfail_isd,aqws).EQ.0 .AND. iwsi.EQ.0) then
c ---    Calm here based on speed (integer zero)
         ws=0.0
         wd=0.0
      elseif(aqwt.NE.'9') then
c ---    Process numerical values and quality code
         if(INDEX(cfail_isd,aqwd).EQ.0 .AND. iwdi.LT.iwdmiss_isd .AND.
     &      INDEX(cfail_isd,aqws).EQ.0 .AND. iwsi.LT.iwsmiss_isd) then
c ---       Valid WS & WD
            ws=FLOAT(iwsi)/FLOAT(iwsfac_isd)
            wd=FLOAT(iwdi)/FLOAT(iwdfac_isd)
c ---       Adjust speed (conditional)
            if(cwsadj(k).EQ.'ASOSBIAS') then
               iwskt=NINT(ws/fkt2mps)
               ws=(FLOAT(iwskt)+add_kt)*fkt2mps
            endif
         endif
      else
c ---    MISSING wind
      endif
      
c --- Temperature and RH
      if(INDEX(cfail_isd,aqtk).EQ.0 .AND. itki.LT.itmiss_isd) then
         tk=FLOAT(itki)/FLOAT(itfac_isd)+273.15
c ---    Set freeze indicator for precip
         ipctk=20
         if(tk.GE.273.15) ipctk=1
c ---    Dew-point Temperature and calc RH
c ---    J. Bosen (1958) RH equation:  degrees F
         if(INDEX(cfail_isd,aqdp).EQ.0 .AND. idpi.LT.itmiss_isd) then
            dpf=(FLOAT(idpi)/FLOAT(itfac_isd))*1.8+32.
            tf=(tk-273.15)*1.8+32.
            irh=NINT(100.*((173.-0.1*tf+dpf)/(173.+0.9*tf))**8)
c ---       Condition range 1-100%
            irh=MAX(irh,1)
            irh=MIN(irh,100)
         endif

         if(ldb) write(iolst,*)'TK,IPCTK = ',tk,ipctk

      endif

c --- Sea-level pressure
      if(INDEX(cfail_isd,aqslp).EQ.0 .AND. islpi.LT.islpmiss_isd)
     &   slp=FLOAT(islpi)/FLOAT(islpfac_isd)

c --- Ceiling height
      if(INDEX(cfail_isd,aqceil3(1:1)).EQ.0 .AND.
     &         iceili.LT.iceilmiss_isd) then
         iceil=iceili/iceilfac_isd
         if(iceil.EQ.22000) then
c ---       Unlimited (m):  replace with code
            iceil=999
         else
c ---       Change meters to 100s of feet
            iceil=NINT(0.01*iceil/0.3048)
         endif
      endif

c --- ADDITIONAL Data Section
c ---------------------------

c --- There should be a 3-char code after character 105
      nlast=LEN_TRIM(aline)
      if(nlast.GT.105) then
         if(aline(106:108).EQ.'ADD') then
c ---       Process the additional variable section when present
            i1=109
            i2=111
            do while (i1.LT.nlast)
               char3=aline(i1:i2)
c ---          Step out if another section is reached
               if(char3.EQ.'REM') EXIT
               if(char3.EQ.'EQD') EXIT
               if(char3.EQ.'QNN') EXIT
c ---          Match to list of ADD elements
               imatch=0
               do i=1,nadd_isd
                  n=3
                  if(code_add(i)(3:3).EQ.'_') n=2
                  if(char3(1:n).EQ.code_add(i)(1:n)) then
                     imatch=1
c ---                Use this?
                     if(iuse_add(i).EQ.0) then
c ---                   SKIP to next element
                        i1=i2+length_add(i)+1
                        i2=i1+2
                     else
c ---                   Process selected elements

c ---                   (AA) Precip Rate & IPCODE
                        if(char3(1:2).EQ.'AA') then
                           read(aline(i2+1:i2+length_add(i)),
     &                          '(i2,i4,1x,a1)') nhr,ipri,aqflg1
                           if(INDEX(cfail_isd,aqflg1).EQ.0 .AND.
     &                        ipri.LT.9999) then
c ---                         Keep 1-hour accumulation or any zero
                              if(ipri.EQ.0 .OR. nhr.EQ.1) then
                                 prate=0.1*FLOAT(ipri)
                                 if(ipri.EQ.0) then
                                    ipcode=0
                                 elseif(tk.LT.rmiss4) then
                                    ipcode=ipctk
                                 endif
                              endif
                           endif
                           i1=i2+length_add(i)+1
                           i2=i1+2

         if(ldb) write(iolst,*)'AA: PRATE,IPCODE = ',prate,ipcode


c ---                   (GA) Cloud Cover (method b: report by layer)
                        elseif(char3(1:2).EQ.'GA') then
c ---                      Select max cloud cover
                           read(aline(i2+1:i2+length_add(i)),
     &                          '(i2,a1,10x)') icci,aqflg1
                           if(INDEX(cfail_isd,aqflg1).EQ.0 .AND.
     &                        icci.LE.10) then
c ---                         Coverage (octas) (Do not use codes >10)
                              if(iccb.EQ.imiss4) then
                                 iccb=icldmap11(icci+1)
                              else
                                 iccb=MAX(iccb,icldmap11(icci+1))
                              endif
                           endif
                           i1=i2+length_add(i)+1
                           i2=i1+2

c ---                   (GD) Cloud Cover (method a: summation report)
                        elseif(char3(1:2).EQ.'GD') then
c ---                      Keep only last entry (overwrite previous)
                           read(aline(i2+1:i2+length_add(i)),
     &                          '(i1,i2,a1,8x)') icci,icc2i,aqflg1
                           if(INDEX(cfail_isd,aqflg1).EQ.0 .AND.
     &                        icc2i.LE.10) then
c ---                         Use coverage code #2 (octas)
c ---                         (Do not use codes >10 here)
                              icca=icldmap11(icc2i+1)
                           elseif(INDEX(cfail_isd,aqflg1).EQ.0 .AND.
     &                        icci.LE.6) then
c ---                         Use coverage code #1 (qualitative)
c ---                         (Do not use codes >6 here)
                              icca=icldmap7(icci+1)
                           endif
                           i1=i2+length_add(i)+1
                           i2=i1+2

c ---                   (GF) Cloud Cover (total/opaque)
                        elseif(char3(1:2).EQ.'GF') then
                           read(aline(i2+1:i2+length_add(i)),
     &                          '(2i2,a1,18x)') icci,icc2i,aqflg1
c ---                      Opaque cover (octas) (Do not use codes >10)
                           if(INDEX(cfail_isd,aqflg1).EQ.0 .AND.
     &                        icc2i.LE.10) icco=icldmap11(icc2i+1)
c ---                      Total cover (octas) (Do not use codes >10)
                           if(INDEX(cfail_isd,aqflg1).EQ.0 .AND.
     &                        icci.LE.10) icct=icldmap11(icci+1)
                           i1=i2+length_add(i)+1
                           i2=i1+2

c ---                   (MA) Pressure (station and/or altimeter)
                        elseif(char3(1:2).EQ.'MA') then
                           read(aline(i2+1:i2+length_add(i)),
     &                          '(2(i5,a1))') ipalt,aqflg1,ipstn,aqflg2
                           if(INDEX(cfail_isd,aqflg2).EQ.0 .AND.
     &                        ipstn.LT.99999) then
c ---                         Station pressure is available
                              pres=0.1*FLOAT(ipstn)
                           elseif(INDEX(cfail_isd,aqflg1).EQ.0 .AND.
     &                        ipalt.LT.99999 .AND. zstn.LT.rmiss4) then
c ---                            Compute station pressure from altimeter
c--------------------------------
c --- METHOD A
c ---                            Invert NOAA altimeter formula and drop
c ---                            the 0.3mb offset per ASOS procedures
c ---   [http://www.srh.noaa.gov/images/epz/wxcalc/altimeterSetting.pdf]
c ---   [http://wahiduddin.net/calc/refs/ASOS_Pressure.htm]
c ---                            Ps=[ Palt **a - 8.4229e-05 H]**(1/a)
c ---                            a=.190284, 1/a=5.2553
c *** using B                    pres=0.1*FLOAT(ipalt)
c ***                            pres=(pres**.190284 -
c ***&                                  zstn*8.4229e-05)**5.2553
c--------------------------------
c --- METHOD B
c ---                            Match SMERGE altimeter formula stated
c ---                            to be based on an operational algorithm
c ---                            from the AES Barometry Program of
c ---                            Environment Canada
c ---                            Ps=[ Palt **a - 8.41717e-05 H]**(1/a)
c ---                            a=.19026, 1/a=5.25593
                                 pres=0.1*FLOAT(ipalt)
                                 pres=(pres**.19026 -
     &                                  zstn*8.41717e-05)**5.25593
c--------------------------------
                           endif
                           i1=i2+length_add(i)+1
                           i2=i1+2

c ---                   (AW) Present Weather (automated)
c ---                    Use to refine IPCODE
                        elseif(char3(1:2).EQ.'AW') then
                           read(aline(i2+1:i2+length_add(i)),
     &                          '(i2,a1)') ipw,aqflg1
                           if(INDEX(cfail_isd,aqflg1).EQ.0 .AND.
     &                        ipw.LE.99) then
                              ipc=ipca(ipw+1)
                              if(ipc.EQ.0) then
c ---                            Apply if measured is trace or missing
                                 if(prate.LT.0.01 .OR.
     &                              prate.GT.(rmiss4-1.0)) then
                                    prate=0.0
                                    ipcode=0
                                 endif
                              elseif(ipc.EQ.1) then
                                 ipcode=1
                              elseif(ipc.GT.1 .AND. ipcode.NE.1) then
                                 ipcode=ipc
                              else
c ---                            Indeterminate: use temperature but
c ---                            do not override previous rain result
                                 if(tk.LT.rmiss4 .AND. ipcode.NE.1)
     &                              ipcode=ipctk
                              endif
                           endif
                           i1=i2+length_add(i)+1
                           i2=i1+2

c ---                   (MW) Present Weather (manual)
c ---                    Use to refine IPCODE
                        elseif(char3(1:2).EQ.'MW') then
                           read(aline(i2+1:i2+length_add(i)),
     &                          '(i2,a1)') ipw,aqflg1
                           if(INDEX(cfail_isd,aqflg1).EQ.0 .AND.
     &                        ipw.LE.99) then
                              ipc=ipcm(ipw+1)
                              if(ipc.EQ.0) then
c ---                            Apply if measured is trace or missing
                                 if(prate.LT.0.01 .OR.
     &                              prate.GT.(rmiss4-1.0)) then
                                    prate=0.0
                                    ipcode=0
                                 endif
                              elseif(ipc.EQ.1) then
                                 ipcode=1
                              elseif(ipc.GT.1 .AND. ipcode.NE.1) then
                                 ipcode=ipc
                              else
c ---                            Indeterminate: use temperature but
c ---                            do not override previous rain result
                                 if(tk.LT.rmiss4 .AND. ipcode.NE.1)
     &                              ipcode=ipctk
                              endif
                           endif
                           i1=i2+length_add(i)+1
                           i2=i1+2

         if(ldb) write(iolst,*)'MW: IPW,IPC,IPCODE = ',ipw,ipc,ipcode

                        endif
                     endif
                     EXIT
                  endif
               enddo

c ---          Should have a match by now
               if(imatch.EQ.0) then
                  write(iolst,*)'ERROR in READ_ISD:  no match for '//
     &                          'ADD element named: '//char3
                  write(iolst,*) TRIM(aline)
                  stop 'HALT in READ_ISD -- see list file'
               endif
            enddo

c ---       Finish summation cloud-cover selection
            if(icca.LT.imiss4) then
               iccs=icca
            elseif(iccb.LT.imiss4) then
               iccs=iccb
            endif

         if(ldb) write(iolst,*)'Done: PRATE,IPCODE  = ',prate,ipcode
         if(ldb) write(iolst,*)'ICCO,ICCT,ICCA,ICCB = ',
     &                          icco,icct,icca,iccb


         endif
      endif


      return

999   write(iolst,*)'ERROR in READ_ISD:  unknown quality flag = ',chk
      write(iolst,*)'Known flags are = '//cqfmand_isd
      write(iolst,*) TRIM(aline)
      stop 'HALT in READ_ISD -- see list file'

      end

c-----------------------------------------------------------------------
      subroutine read_asos1(io,io2,iolst,k,
     &                      iyr,imo,idy,jul,ihr,isec,ieof,
     &                      ws,wd,iceil,icco,icct,iccs,tk,irh,pres,
     &                      ipcode,prate,iobs)
c-----------------------------------------------------------------------
c
c --- SURFGEN   Version: 7.0.0          Level: 150330        READ_ASOS1
c               D. Strimaitis, Exponent Inc.
c
c --- PURPOSE:  Extracts data from 1 record of pair of ASOS1-format
c               6405 and 6406 files
c
c --- INPUT:
c                  IO - integer   - File unit number for 6405 file
c                 IO2 - integer   - File unit number for 6406 file
c               IOLST - integer   - Unit number of list file
c                   K - integer   - Current station (file) index
c
c     Module CONST
c            izero, imiss4, rmiss4
c     Module USERINF
c            cstaid(),cwsadj(),isectzadj(),isec2utc()
c
c --- OUTPUT:
c
c ---          Date-Time is END of observation period
c                 IYR - integer   - Year (YYYY)
c                 IMO - integer   - Month (MM)
c                 IDY - integer   - Day (DD)
c                 JUL - integer   - Julian Day (DDD)
c                 IHR - integer   - Hour (hh)
c                ISEC - integer   - Second (ssss)
c
c                IEOF - integer   - Completion flag (end of file)
c                  WS - real      - Wind speed (m/s)
c                  WD - real      - Wind direction
c               ICEIL - integer   - Ceiling height (100s feet)
c                ICCO - integer   - Opaque cloud cover
c                ICCT - integer   - Total cloud cover
c                ICCS - integer   - Summation cloud cover
c                  TK - real      - Air temperature (K)
c                 IRH - integer   - Relative humidity (%)
c                PRES - real      - Station pressure  (mb)
c              IPCODE - integer   - Precipitation code
c                                      0 = No precip
c                                      1 = Liquid precip
c                                     20 = Frozen precip
c                                   9999 = Missing
c               PRATE - real      - Liquid precip rate (mm/hr)
c                IOBS - integer   - Observation flag
c                                   -1: drop observation
c                                    0: regular observation
c                                    1: special observation
c
c --- READ_ASOS1 called by: SETUP, COMP
c --- READ_ASOS1 calls:     CAL_DAY, TIME_INCR, MK_CASE, ADD_DLIM
c                           GETCOLS
c----------------------------------------------------------------------
c --- Notes
c 1.  All observations in the output period are processed.
c 2.  Observations are flagged as regular.
c 3.  Calling program determines if values returned overwrite previous
c     values found for the period
c 4.  Data not contained in the 6405 an 6406 datasets are set missing.
c 5.  Precipitation accumulated over 15 minutes is reported each minute
c     so the output is scaled up to an hourly rate.
c----------------------------------------------------------------------
c --- Access modules
      use const
      use userinf

      implicit none

      integer        :: mxcol
      parameter (mxcol=16)

c --- Declare arguments
      integer        :: io,io2,iolst,k
      integer        :: iyr,imo,idy,jul,ihr,isec
      integer        :: ieof,iceil,icco,icct,iccs,irh,ipcode,iobs
      real           :: ws,wd,tk,pres,prate

c --- Locals
      logical        :: ldb
      character(len=320) :: aline6405,aline6406
      character(len=40)  :: acol6405(mxcol),acol6406(mxcol)
      character(len=16)  :: atime6405,atime6406
      character(len=10)  :: aid10
      integer        :: i,il,ir,npress
      integer        :: iwskt,iwd,itf,itdpf
      integer        :: imm,ihrmmutc
      integer        :: ncol6405,ncol6406
      real           :: pstn(3),fin2mb,fkt2mps,add_kt

      data ldb/.TRUE./

c --- Wind speed conversion and ASOS bias adjustment
      data fkt2mps/0.5144/, add_kt/0.5/

c --- Pressure conversion to mb
      data fin2mb/33.86389/

c --- Initialize all output variables to missing
      ws=rmiss4
      wd=rmiss4
      tk=rmiss4
      irh=imiss4
      pres=rmiss4
      iceil=imiss4
      icco=imiss4
      icct=imiss4
      iccs=imiss4
      ipcode=imiss4
      prate=rmiss4
c --- Impose observation type
      iobs=0
c --- Initialize selected locals to missing
      pstn=rmiss4
      itf=imiss4
      itdpf=imiss4

c --- Read 1 record from 6405 file
      ieof=0
c --- Rely on padding to initialize full string to blanks
      aline6405=' '
      read(io,'(a320)',iostat=ieof) aline6405
      if(ieof.NE.0) return
c --- Force upper case
      call MK_CASE(iolst,'READ_ASOS1','UPPER',aline6405)
c --- Check station ID (allow 10 characters - trailing blanks)
      aid10='          '
      aid10(1:9)=aline6405(1:9)
      if(INDEX(aid10,cstaid(k)).NE.0) then
         write(iolst,*)'ERROR in READ_ASOS1:  Station ID not found'
         write(iolst,*)'Requested ID = ',cstaid(k)
         write(iolst,*)'ID in file   = ',aid10
         write(iolst,*) TRIM(aline6405)
         stop 'HALT in READ_ASOS1 -- see list file'
      endif
c --- Extract date-time
      atime6405=aline6405(14:29)
      read(atime6405(1:4),'(i4)') iyr
      read(atime6405(5:6),'(i2)') imo
      read(atime6405(7:8),'(i2)') idy
      read(atime6405(9:10),'(i2)') ihr
      read(atime6405(11:12),'(i2)') imm
      read(atime6405(13:16),'(i4)') ihrmmutc
      isec=imm*60
      call CAL_DAY(iolst,'READ_ASOS1','TOJULIAN',iyr,imo,idy,jul)
c --- Adjust date to UTC and verify UTC time
      call TIME_INCR(iolst,'READ_ASOS1',iyr,jul,ihr,isec,izero,
     &               isec2utc(k))
      if((ihr*100+imm).NE.ihrmmutc) then
         write(iolst,*)'ERROR in READ_ASOS1:  UTC times do not match'
         write(iolst,*)'LST date and time read = ',atime6405(1:12)
         write(iolst,*)'         UTC time read = ',atime6405(13:16)
         write(iolst,*)'  Expect LST time zone = ',citzone(k)
         write(iolst,*)' UTC computed from LST = ',(ihr*100+imm)
         stop 'HALT in READ_ASOS1 -- see list file'
      endif

c --- Adjust time to output time zone from UTC
      call TIME_INCR(iolst,'READ_ASOS1',iyr,jul,ihr,isec,izero,
     &              (isectzadj(k)-isec2utc(k)))
      call CAL_DAY(iolst,'READ_ASOS1','TOGREGOR',iyr,imo,idy,jul)

c --- Read 1 record from 6406 file
      ieof=0
c --- Rely on padding to initialize full string to blanks
      aline6406=' '
      read(io2,'(a320)',iostat=ieof) aline6406
      if(ieof.NE.0) return
c --- Force upper case
      call MK_CASE(iolst,'READ_ASOS1','UPPER',aline6406)
c --- Expect record to match 6405 record station and time
      if(aline6405(1:29).NE.aline6406(1:29)) then
         write(iolst,*)'ERROR in READ_ASOS1:  6405 and 6406 differ'
         write(iolst,*)'Expected the same station and time'
         write(iolst,*)'6405 = '//TRIM(aline6405)
         write(iolst,*)'6406 = '//TRIM(aline6406)
         stop 'HALT in READ_ASOS1 -- see list file'
      endif

c --- Prepare the data records for processing
c --- Remove the [] bracketed field (when present) from the 6406 record
      il=INDEX(aline6406,'[')
      ir=INDEX(aline6406,']')
      if(il.GT.0 .AND. ir.GT.il) then
         do i=il,ir
            aline6406(i:i)=' '
         enddo
      endif
c --- Parse record into columns (may not always be fixed format)
      call ADD_DLIM(',',aline6405)
      call ADD_DLIM(',',aline6406)
c --- Break out columns into strings
c --- Initialize string array to blanks
      acol6405=' '
      acol6406=' '
      call GETCOLS(',',aline6405,mxcol,ncol6405,acol6405)
      call GETCOLS(',',aline6406,mxcol,ncol6406,acol6406)

c      if(LDB) then
c         write(iolst,*)'aline6405= '//aline6405
c         write(iolst,*)'aline6406= '//aline6406
c         write(iolst,*)'acol6405= ',(acol6405(i),i=1,ncol6405)
c         write(iolst,*)'acol6406= ',(acol6406(i),i=1,ncol6406)
c      endif

c --- Process data on record only if expected number of columns found
c --- 6405
c ---------
      if(ncol6405.EQ.13) then
c ---    Wind direction (9)
         if(INDEX(acol6405(9),'M').EQ.0) then
            read(acol6405(9),'(i)') iwd
            wd=FLOAT(iwd)
         endif
c ---    Wind speed (10)
         if(INDEX(acol6405(10),'M').EQ.0) then
            read(acol6405(10),'(i)') iwskt
c ---       Adjust speed (conditional)
            if(cwsadj(k).EQ.'ASOSBIAS') then
               ws=(FLOAT(iwskt)+add_kt)*fkt2mps
            else
               ws=FLOAT(iwskt)*fkt2mps
            endif
         endif
      endif
c --- 6406
c ---------
      if(ncol6406.EQ.10) then
c ---    Temperature (9)
         if(INDEX(acol6406(9),'M').EQ.0) then
            read(acol6406(9),'(i)') itf
            tk=273.15+FLOAT(itf-32)/1.8
         endif
c ---    Dew-point temperature (10)
         if(INDEX(acol6406(10),'M').EQ.0) then
            read(acol6406(10),'(i)') itdpf
         endif
c ---    Station pressures (6,7,8)
         npress=0
         do i=6,8
            if(INDEX(acol6406(i),'M').EQ.0) then
               read(acol6406(i),'(f)') pstn(i-5)
               npress=npress+1
            endif
         enddo
c ---    Precip type (after temperature is obtained)
         if(INDEX(acol6406(3),'NP').GT.0) then
            ipcode=0
            prate=0.0
         elseif(INDEX(acol6406(3),'R').GT.0) then
            ipcode=1
         elseif(INDEX(acol6406(3),'S').GT.0) then
            ipcode=20
         elseif(INDEX(acol6406(3),'M').EQ.0) then
c ---       Indeterminate, use temperature if available
            if(itf.LT.imiss4) then
               if(itf.GE.32) then
                  ipcode=1
               else
                  ipcode=20
               endif
            endif
         endif
c ---    Construct derivatives
c ---    Relative humidity
         if(itf.LT.imiss4 .AND. itdpf.LT.imiss4) then
c ---       J. Bosen (1958) RH equation:  degrees F
            irh=NINT(100.*((173.-0.1*FLOAT(itf)+FLOAT(itdpf))/
     &                     (173.+0.9*FLOAT(itf)))**8)
c ---       Condition range 1-100%
            irh=MAX(irh,1)
            irh=MIN(irh,100)
         endif
c ---    Station pressure (average)
         if(npress.GT.0) then
            pres=0.0
            do i=1,3
               if(pstn(i).LT.(rmiss4-1.0)) pres=pres+pstn(i)
            enddo
            pres=fin2mb*(pres/FLOAT(npress))
         endif

      endif

      if(ldb) then
         write(iolst,*)'READ_ASOS1: data for period '//atime6405
         write(iolst,*)'WS,WD,TK,IRH      = ',ws,wd,tk,irh
         write(iolst,*)'PRES,PRATE,IPCODE = ',pres,prate,ipcode
      endif

      return
      end

c----------------------------------------------------------------------
      subroutine fmtdate(iolst,cfrom,action,c10date,iyr,imon,iday)
c----------------------------------------------------------------------
c
c --- SURFGEN   Version: 7.0.0          Level: 150330         FMT_TIME
c               D. Strimaitis, Exponent Inc.
c
c --- PURPOSE:  Convert between date string and date integers
c
c --- INPUTS:
c               IOLST - integer   - Unit number of output list file
c               CFROM - char*     - Called-From string to report error
c              ACTION - char*     - Action to perform
c                                   SPLIT returns year, month, day
c                                   GROUP returns 'YYYY-MM-DD'
c
c --- INPUT/OUTPUT:
c             c10date - char*10   - Date (YYYY-MM-DD)
c                 IYR - integer   - Year (YYYY)
c                IMON - integer   - Month (MM)
c                IDAY - integer   - Day (DD)
c
c --- FMTDATE called by:  ()
c --- FMTDATE calls:      none
c----------------------------------------------------------------------
      implicit none

c --- Declare arguments
      character(len=*)  :: action, cfrom
      character(len=10) :: c10date
      integer           :: iolst,iyr,imon,iday

c --- Locals
      logical  :: lbad

      if(action.EQ.'SPLIT') then
c ---    Convert from string
c ---    QA format of string (Length=10 with 2 dashes YYYY-MM-DD)
         lbad=.FALSE.
         if(LEN_TRIM(c10date).NE.10) lbad=.TRUE.
         if(c10date(5:5).NE.'-') lbad=.TRUE.
         if(c10date(8:8).NE.'-') lbad=.TRUE.
         if(lbad) then
            write(iolst,*)'ERROR in FMTDATE:  Invalid date format'
            write(iolst,*)' --- Expected format YYYY-MM-DD'
            write(iolst,*)' --- Found           ',c10date
            write(iolst,*)'Called from  '//cfrom
            write(*,*)'Halted in FMTDATE -- see list file'
            stop
         endif
c ---    Read integer fields from string
         read(c10date,'(i4.4,1x,i2.2,1x,i2.2)') iyr,imon,iday  
      elseif(action.EQ.'GROUP') then
c ---    Write integer fields to string
         write(c10date,'(i4.4,a1,i2.2,a1,i2.2)') iyr,'-',imon,'-',iday  
      else
         write(iolst,*)'ERROR in FMTDATE -- Invalid ACTION'
         write(iolst,*)'   ACTION expected: SPLIT or GROUP'
         write(iolst,*)'   ACTION found   : ',action
         write(iolst,*)'Called from  '//cfrom
         write(*,*)'Halted in FMTDATE -- see list file'
         stop
      endif

      return
      end


c----------------------------------------------------------------------
      subroutine fmttime(iolst,cfrom,action,c8time,ihr,isec)
c----------------------------------------------------------------------
c
c --- SURFGEN   Version: 7.0.0          Level: 150330         FMT_TIME
c               D. Strimaitis, Exponent Inc.
c
c --- PURPOSE:  Convert between time string and time integers
c
c --- INPUTS:
c               IOLST - integer   - Unit number of output list file
c               CFROM - char*     - Called-From string to report error
c              ACTION - char*     - Action to perform
c                                   SPLIT returns HH, SSSS
c                                   GROUP returns 'HH:MM:SS'
c
c --- INPUT/OUTPUT:
c              c8time - char*8    - Time (HH:MM:SS)
c                 IHR - integer   - Hours (HH)
c                ISEC - integer   - Seconds (SSSS)
c
c --- FMTTIME called by:  ()
c --- FMTTIME calls:      none
c----------------------------------------------------------------------
      implicit none

c --- Declare arguments
      character(len=*)  :: action, cfrom
      character(len=8)  :: c8time
      integer           :: iolst,ihr,isec

c --- Locals
      logical  :: lbad
      integer  :: imm,iss

      if(action.EQ.'SPLIT') then
c ---    Convert from string
c ---    QA format of string (Length=8 with 2 colons HH:MM:SS)
         lbad=.FALSE.
         if(LEN_TRIM(c8time).NE.8) lbad=.TRUE.
         if(c8time(3:3).NE.':') lbad=.TRUE.
         if(c8time(6:6).NE.':') lbad=.TRUE.
         if(lbad) then
            write(iolst,*)'ERROR in FMTTIME:  Invalid time format'
            write(iolst,*)' --- Expected format HH:MM:SS'
            write(iolst,*)' --- Found           ',c8time
            write(iolst,*)'Called from  '//cfrom
            write(*,*)'Halted in FMTTIME -- see list file'
            stop
         endif
c ---    Read integer fields from string
         read(c8time,'(i2.2,1x,i2.2,1x,i2.2)') ihr,imm,iss
         isec=iss+imm*60
      elseif(action.EQ.'GROUP') then
c ---    Write integer fields to string
         imm=isec/60
         iss=isec-imm*60
         write(c8time,'(i2.2,a1,i2.2,a1,i2.2)') ihr,':',imm,':',iss  
      else
         write(iolst,*)'ERROR in FMTTIME -- Invalid ACTION'
         write(iolst,*)'   ACTION expected: SPLIT or GROUP'
         write(iolst,*)'   ACTION found   : ',action
         write(*,*)'Halted in FMTTIME -- see list file'
         write(iolst,*)'Called from  '//cfrom
         stop
      endif

      return
      end

c-----------------------------------------------------------------------
      subroutine move_dat(iolst,action,kf,ifrom,ito)
c-----------------------------------------------------------------------
c
c --- SURFGEN   Version: 7.0.0          Level: 150330          MOVE_DAT
c               D. Strimaitis, Exponent Inc.
c
c --- PURPOSE:  Move data in arrays for station KF from element IFROM
c               to element ITO subject to ACTION requested
c
c --- INPUT:
c               IOLST - integer   - List file unit number
c              ACTION - char*     - Action to perform
c                                   DATETIME (replace timestamps)
c                                   REPLACE (complete overwrite)
c                                   UPDATE (with non-missing values)
c                                   FILLMISS (replace missing values)
c                                   MISSING (make all missing)
c                  KF - integer   - File index (station)
c               IFROM - integer   - Array index to copy FROM
c                                   (0 if action=MISSING)
c                 ITO - integer   - Array index to copy TO
c
c     Module CONST
c            rmiss4, imiss4
c
c --- INPUT/OUTPUT:
c     Module DATA
c            i8datim1(),i8datim2()
c            ws(),wd(),tk(),pres(),prate()
c            iceil(),icco(),icct(),iccs(),irh(),ipcode()
c
c --- MOVE_DAT called by: SETUP, COMP
c --- MOVE_DAT calls:     (none)
c----------------------------------------------------------------------
c --- Access module data
      use const
      use data

      implicit none

c --- Declare arguments
      character (len=*) :: action
      integer           :: iolst,kf,ifrom,ito

c --- Locals
      integer           :: ival,iuse,inew,iold
      real              :: rval

      if(ito.LT.1 .OR. ito.GT.3) then
         write(iolst,*) 'ERROR in MOVE_DAT:  invalid TO index = ',ito
         write(iolst,*) 'Expected index = 1, 2 or 3'
         stop 'HALT in MOVE_DAT -- see list file'
      endif
      if(ifrom.LT.0 .OR. ifrom.GT.3) then
         write(iolst,*) 'ERROR in MOVE_DAT:  invalid FROM index = ',
     &                   ifrom
         write(iolst,*) 'Expected index = 0, 1, 2 or 3'
         stop 'HALT in MOVE_DAT -- see list file'
      endif

      SELECT CASE (action)

      CASE ("DATETIME")
c ---    Copy (overwrite) date-times
         i8datim1(ito,kf)= i8datim1(ifrom,kf)
         i8datim2(ito,kf)= i8datim2(ifrom,kf)

      CASE ("MISSING")
c ---    Initialize to missing values
         ws(ito,kf)      = rmiss4     
         wd(ito,kf)      = rmiss4
         iceil(ito,kf)   = imiss4
         icco(ito,kf)    = imiss4
         icct(ito,kf)    = imiss4
         iccs(ito,kf)    = imiss4
         tk(ito,kf)      = rmiss4
         irh(ito,kf)     = imiss4
         pres(ito,kf)    = rmiss4
         ipcode(ito,kf)  = imiss4
         prate(ito,kf)   = rmiss4

      CASE ("REPLACE")
c ---    Copy (overwrite) everything
         ws(ito,kf)      = ws(ifrom,kf)     
         wd(ito,kf)      = wd(ifrom,kf)     
         iceil(ito,kf)   = iceil(ifrom,kf)  
         icco(ito,kf)    = icco(ifrom,kf)    
         icct(ito,kf)    = icct(ifrom,kf)    
         iccs(ito,kf)    = iccs(ifrom,kf)    
         tk(ito,kf)      = tk(ifrom,kf)     
         irh(ito,kf)     = irh(ifrom,kf)    
         pres(ito,kf)    = pres(ifrom,kf)   
         ipcode(ito,kf)  = ipcode(ifrom,kf) 
         prate(ito,kf)   = prate(ifrom,kf)  

      CASE ("UPDATE")
c ---    UPDATE with non-missing values
c ---    Overwrite values only if replacements are non-missing
         ival=imiss4-1
         rval=rmiss4-1.0
         if(ws(ifrom,kf)    .LT.rval)     ws(ito,kf)= ws(ifrom,kf)
         if(wd(ifrom,kf)    .LT.rval)     wd(ito,kf)= wd(ifrom,kf)
         if(iceil(ifrom,kf) .LT.ival)  iceil(ito,kf)= iceil(ifrom,kf)
         if(icco(ifrom,kf)  .LT.ival)   icco(ito,kf)= icco(ifrom,kf)
         if(icct(ifrom,kf)  .LT.ival)   icct(ito,kf)= icct(ifrom,kf)
         if(iccs(ifrom,kf)  .LT.ival)   iccs(ito,kf)= iccs(ifrom,kf)
         if(tk(ifrom,kf)    .LT.rval)     tk(ito,kf)= tk(ifrom,kf)
         if(irh(ifrom,kf)   .LT.ival)    irh(ito,kf)= irh(ifrom,kf)
         if(pres(ifrom,kf)  .LT.rval)   pres(ito,kf)= pres(ifrom,kf)
         if(ipcode(ifrom,kf).LT.ival) ipcode(ito,kf)= ipcode(ifrom,kf)
         if(prate(ifrom,kf) .LT.rval)  prate(ito,kf)= prate(ifrom,kf)

      CASE ("FILLMISS")
c ---    Update missing values with non-missing values
         ival=imiss4-1
         rval=rmiss4-1.0
         if(ws(ifrom,kf)    .LT.rval .AND.
     &      ws(ito,kf)      .GT.rval)     ws(ito,kf)= ws(ifrom,kf)
         if(wd(ifrom,kf)    .LT.rval .AND.
     &      wd(ito,kf)      .GT.rval)     wd(ito,kf)= wd(ifrom,kf)
         if(iceil(ifrom,kf) .LT.ival .AND.
     &      iceil(ito,kf)   .GT.ival)  iceil(ito,kf)= iceil(ifrom,kf)
         if(icco(ifrom,kf)  .LT.ival .AND.
     &      icco(ito,kf)    .GT.ival)   icco(ito,kf)= icco(ifrom,kf)
         if(icct(ifrom,kf)  .LT.ival .AND.
     &      icct(ito,kf)    .GT.ival)   icct(ito,kf)= icct(ifrom,kf)
         if(iccs(ifrom,kf)  .LT.ival .AND.
     &      iccs(ito,kf)    .GT.ival)   iccs(ito,kf)= iccs(ifrom,kf)
         if(tk(ifrom,kf)    .LT.rval .AND.
     &      tk(ito,kf)      .GT.rval)     tk(ito,kf)= tk(ifrom,kf)
         if(irh(ifrom,kf)   .LT.ival .AND.
     &      irh(ito,kf)     .GT.ival)    irh(ito,kf)= irh(ifrom,kf)
         if(pres(ifrom,kf)  .LT.rval .AND.
     &      pres(ito,kf)    .GT.rval)   pres(ito,kf)= pres(ifrom,kf)
         if(ipcode(ifrom,kf).LT.ival .AND.
     &      ipcode(ito,kf)  .GT.ival) ipcode(ito,kf)= ipcode(ifrom,kf)
         if(prate(ifrom,kf) .LT.rval .AND.
     &      prate(ito,kf)   .GT.rval)  prate(ito,kf)= prate(ifrom,kf)

      CASE DEFAULT
c ---    Bad calling action
         write(iolst,*)
         write(iolst,*)'ERROR in MOVE_DAT:  Unknown ACTION = ',action
         write(iolst,*)'Expected: DATETIME, REPLACE, UPDATE, '//
     &                 'FILLMISS, or MISSING'
         stop 'HALTED in MOVE_DAT -- see list file'

      END SELECT


      return
      end

c-----------------------------------------------------------------------
      subroutine finish(iolst)
c-----------------------------------------------------------------------
c
c --- SURFGEN   Version: 7.0.0          Level: 150330            FINISH
c               D. Strimaitis, Exponent Inc.
c
c --- PURPOSE:  Report final information to list file
c
c --- INPUT:
c               IOLST - integer   - List file unit number
c
c     Module COUNT
c            npdproc,nvar,nmiss(),ncalm(),cvar()
c     Module USERINF
c            nfiles
c
c --- OUTPUT:
c
c --- FINISH called by: (main)
c --- FINISH calls:     (none)
c----------------------------------------------------------------------
c --- Access module data
      use counts
      use userinf

      implicit none

c --- Declare arguments
      integer           :: iolst

c --- Locals
      integer           :: i,k

c --- Report total number of periods and missing counts
      write(iolst,*)
      write(iolst,*)'Number of time-periods written:  ',npdproc
      write(iolst,*)
      write(iolst,*)'Number of time-periods with all stations missing:'
      write(iolst,'(100a20)') '                    ',cvar
      write(iolst,'(20x,100(i10,10x))') (nmiss(i,nfiles+1),i=1,nvar)
      write(iolst,*)
      write(iolst,*)'Number of Missings for Each Station:'
      write(iolst,'(100a20)') '                    ',cvar
      do k=1,nfiles
         write(iolst,'(a20,100(i10,10x))') cstaid(k),
     &                                   (nmiss(i,k),i=1,nvar)
      enddo
      write(iolst,*)
      write(iolst,*)'Number of Calms for Each Station:'
      write(iolst,'(a20,a10)') '                    ','Calm'
      do k=1,nfiles
         write(iolst,'(a20,i10)') cstaid(k),ncalm(k)
      enddo

      write(iolst,*)
      write(iolst,*)

      return
      end
c----------------------------------------------------------------------
      subroutine add_dlim(dlim,aline)
c----------------------------------------------------------------------
c
c --- SURFGEN   Version: 7.0.0          Level: 150330         ADD_DLIM
c               D. Strimaitis, Exponent Inc.
c
c --- PURPOSE:  Add delimeters to a string to delineate columns.
c               Assumes variable number of spaces separate the data on
c               the line, and replaces the first space in each cluster
c               with the delimeter that is provided.
c
c --- INPUT:
c                DLIM - char*1    - Column delimiter
c               ALINE - char*     - Line of values separated by spaces
c
c --- OUTPUT:
c               ALINE - char*     - Line of values separated by the
c                                   delimiter and remaining spaces
c
c --- ADD_DLIM called by: ()
c --- ADD_DLIM calls:     
c----------------------------------------------------------------------

      implicit none

c --- Declare arguments
      character(len=*)    :: aline
      character(len=1)    :: dlim

c --- Locals 
      logical :: ldlim
      integer :: i,ilen

c --- Left-justify
      aline=ADJUSTL(aline)

c --- Active length of line
      ilen=LEN_TRIM(aline)

c --- Set replace status (ldlim=T means replace space)
      ldlim=.TRUE.

c --- Process each character on line
      do i=1,ilen
         if(aline(i:i).EQ.' ') then
            if(ldlim) then
               aline(i:i)=dlim
               ldlim=.FALSE.
            endif
         else
            ldlim=.TRUE.
         endif
      enddo

      return
      end

c-----------------------------------------------------------------------
      subroutine getcols(dlim,aline,mxcol,ncol,acol)
c-----------------------------------------------------------------------
c
c --- SURFGEN   Version: 7.0.0          Level: 150330           GETCOLS
c               D. Strimaitis, Exponent Inc.
c
c --- PURPOSE:  Split a string with delimiters into a set of strings
c               to extract columns of data
c
c --- INPUT:
c                DLIM - char*1    - Column delimiter
c               ALINE - char*     - Line of values separated by
c                                   delimiter
c               MXCOL - integer   - Maximum number of columns to search
c
c --- OUTPUT:
c                NCOL - integer   - Number of columns found
c              ACOL() - char*     - Array of strings
c
c --- GETCOLS called by: ()
c --- GETCOLS calls:     
c----------------------------------------------------------------------
c --- Calling routine sets up acol() array for specific string
c --- No checking done here to verify column widths fit into array
c----------------------------------------------------------------------

      implicit none

c --- Declare arguments
      integer             :: mxcol, ncol
      character(len=1)    :: dlim
      character(len=*)    :: acol(mxcol)
      character(len=*)    :: aline

c --- Locals 
      integer :: i,k,ilen,i1,i2,num

c --- Active length of input string
      ilen=LEN_TRIM(aline)

      ncol=0
      i1=1
      i2=ilen

c --- Try to fill max number of columns
      do k=1,mxcol
c ---    Mark delimiter position after character i1
         i=INDEX(aline(i1:i2),dlim)
         if(i.EQ.1) then
c ---       Delimiter is adjacent to previous delimiter
c ---       Retain blank field in output array
            ncol=ncol+1
c ---       Reset pointers for next search
            i1=i1+1
            i2=ilen
         elseif(i.GT.0) then
c ---       Delimiter found in range i1 to i2
c ---       Place string to left into output array
            ncol=ncol+1
c ---       Delimiter position 'i' is relative to i1
            num=i-1
            i2=i1+num-1
            acol(ncol)(1:num)=aline(i1:i2)
c ---       Reset pointers for next search
            i1=i2+2
            i2=ilen
         else
c ---       No delimiter, place remaining string into output array
            ncol=ncol+1
            num=i2-i1+1
            acol(ncol)(1:num)=aline(i1:i2)
c ---       Done
            EXIT
         endif
      enddo

      return
      end
